import logging
from collections import defaultdict, deque
from decimal import Decimal
from enum import Enum
from functools import lru_cache
from math import ceil, floor
from typing import Dict, List, Tuple

import pandas as pd
from bidict import bidict

from hummingbot.client.performance import PerformanceMetrics
from hummingbot.client.settings import AllConnectorSettings
from hummingbot.connector.exchange_base import ExchangeBase
from hummingbot.core.clock import Clock
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.limit_order import LimitOrder
from hummingbot.core.data_type.trade_fee import TokenAmount
from hummingbot.core.event.events import (
    BuyOrderCompletedEvent,
    MarketOrderFailureEvent,
    OrderCancelledEvent,
    OrderExpiredEvent,
    OrderFilledEvent,
    SellOrderCompletedEvent,
)
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.strategy.cross_exchange_market_making.cross_exchange_market_making_config_map_pydantic import (
    CrossExchangeMarketMakingConfigMap,
    PassiveOrderRefreshMode,
)
from hummingbot.strategy.maker_taker_market_pair import MakerTakerMarketPair
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
from hummingbot.strategy.strategy_py_base import StrategyPyBase

from .order_id_market_pair_tracker import OrderIDMarketPairTracker

s_float_nan = float("nan")
s_decimal_zero = Decimal(0)
s_decimal_nan = Decimal("nan")
s_logger = None


class LogOption(Enum):
    NULL_ORDER_SIZE = 0
    REMOVING_ORDER = 1
    ADJUST_ORDER = 2
    CREATE_ORDER = 3
    MAKER_ORDER_FILLED = 4
    STATUS_REPORT = 5
    MAKER_ORDER_HEDGED = 6


class CrossExchangeMarketMakingStrategy(StrategyPyBase):

    OPTION_LOG_ALL = (
        LogOption.NULL_ORDER_SIZE,
        LogOption.REMOVING_ORDER,
        LogOption.ADJUST_ORDER,
        LogOption.CREATE_ORDER,
        LogOption.MAKER_ORDER_FILLED,
        LogOption.STATUS_REPORT,
        LogOption.MAKER_ORDER_HEDGED
    )

    ORDER_ADJUST_SAMPLE_INTERVAL = 5
    ORDER_ADJUST_SAMPLE_WINDOW = 12

    SHADOW_MAKER_ORDER_KEEP_ALIVE_DURATION = 60.0 * 15
    CANCEL_EXPIRY_DURATION = 60.0

    @classmethod
    def logger(cls):
        global s_logger
        if s_logger is None:
            s_logger = logging.getLogger(__name__)
        return s_logger

    def init_params(self,
                    config_map: CrossExchangeMarketMakingConfigMap,
                    market_pairs: List[MakerTakerMarketPair],
                    status_report_interval: float = 900,
                    logging_options: int = OPTION_LOG_ALL,
                    hb_app_notification: bool = False
                    ):
        """
        Initializes a cross exchange market making strategy object.

        :param config_map: Strategy configuration map
        :param market_pairs: list of cross exchange market pairs
        :param logging_options: bit field for what types of logging to enable in this strategy object
        :param hb_app_notification:
        """
        self._config_map = config_map
        self._market_pairs = {
            (market_pair.maker.market, market_pair.maker.trading_pair): market_pair
            for market_pair in market_pairs
        }
        self._maker_markets = set([market_pair.maker.market for market_pair in market_pairs])
        self._taker_markets = set([market_pair.taker.market for market_pair in market_pairs])
        self._all_markets_ready = False
        self._conversions_ready = False

        self._anti_hysteresis_timers = {}
        self._order_fill_buy_events = {}
        self._order_fill_sell_events = {}
        self._suggested_price_samples = {}

        self._last_timestamp = 0
        self._status_report_interval = status_report_interval
        self._market_pair_tracker = OrderIDMarketPairTracker()

        # Holds ongoing hedging orders mapped to their respective maker fill trades
        self._ongoing_hedging = bidict()

        self._logging_options = logging_options

        self._last_taker_buy_price = None
        self._last_taker_sell_price = None

        self._main_task = None
        self._gateway_quotes_task = None
        self._hedge_maker_order_tasks = []

        self._last_conv_rates_logged = 0
        self._hb_app_notification = hb_app_notification

        # Holds active maker orders, all its taker orders ever created
        self._maker_to_taker_order_ids = {}
        # Holds active taker orders, and their respective maker orders
        self._taker_to_maker_order_ids = {}
        # Holds hedging trade ids for respective maker orders
        self._maker_to_hedging_trades = {}

        all_markets = list(self._maker_markets | self._taker_markets)

        self.add_markets(all_markets)

    @property
    def order_amount(self):
        return self._config_map.order_amount

    @property
    def min_profitability(self):
        return self._config_map.min_profitability / Decimal("100")

    @property
    def order_size_taker_volume_factor(self):
        return self._config_map.order_size_taker_volume_factor / Decimal("100")

    @property
    def order_size_taker_balance_factor(self):
        return self._config_map.order_size_taker_balance_factor / Decimal("100")

    @property
    def order_size_portfolio_ratio_limit(self):
        return self._config_map.order_size_portfolio_ratio_limit / Decimal("100")

    @property
    def top_depth_tolerance(self):
        return self._config_map.top_depth_tolerance

    @property
    def anti_hysteresis_duration(self):
        return self._config_map.anti_hysteresis_duration

    @property
    def limit_order_min_expiration(self):
        return self._config_map.limit_order_min_expiration

    @property
    def status_report_interval(self):
        return self._status_report_interval

    @property
    def adjust_order_enabled(self):
        return self._config_map.adjust_order_enabled

    @property
    def use_oracle_conversion_rate(self):
        return self._config_map.use_oracle_conversion_rate

    @property
    def taker_to_maker_base_conversion_rate(self):
        return self._config_map.conversion_rate_mode.taker_to_maker_base_conversion_rate

    @property
    def taker_to_maker_quote_conversion_rate(self):
        return self._config_map.taker_to_maker_quote_conversion_rate

    @property
    def slippage_buffer(self):
        return self._config_map.slippage_buffer / Decimal("100")

    @property
    def active_maker_limit_orders(self) -> List[Tuple[ExchangeBase, LimitOrder]]:
        return [(ex, order, order.client_order_id) for ex, order in self._sb_order_tracker.active_limit_orders
                if order.client_order_id in self._maker_to_taker_order_ids.keys()]

    @property
    def cached_limit_orders(self) -> List[Tuple[ExchangeBase, LimitOrder]]:
        return self._sb_order_tracker.shadow_limit_orders

    @property
    def active_maker_bids(self) -> List[Tuple[ExchangeBase, LimitOrder]]:
        return [(market, limit_order) for market, limit_order, order_id in self.active_maker_limit_orders
                if limit_order.is_buy]

    @property
    def active_maker_asks(self) -> List[Tuple[ExchangeBase, LimitOrder]]:
        return [(market, limit_order) for market, limit_order, order_id in self.active_maker_limit_orders
                if not limit_order.is_buy]

    @property
    def active_order_canceling(self):
        return self._config_map.active_order_canceling

    @property
    def adjust_orders_enabled(self):
        return self._config_map.adjust_orders_enabled

    @property
    def gas_to_maker_base_conversion_rate(self):
        return self._config_map.gas_to_maker_base_conversion_rate

    @property
    def logging_options(self) -> int:
        return self._logging_options

    @logging_options.setter
    def logging_options(self, logging_options: Tuple):
        self._logging_options = logging_options

    @property
    def market_info_to_active_orders(self) -> Dict[MarketTradingPairTuple, List[LimitOrder]]:
        return self._sb_order_tracker.market_pair_to_active_orders

    @staticmethod
    @lru_cache(maxsize=10)
    def is_gateway_market(market_info: MarketTradingPairTuple) -> bool:
        return market_info.market.name in AllConnectorSettings.get_gateway_amm_connector_names()

    def get_conversion_rates(self, market_pair: MarketTradingPairTuple):
        quote_pair, quote_rate_source, quote_rate, base_pair, base_rate_source, base_rate, gas_pair, gas_rate_source, \
            gas_rate = self._config_map.conversion_rate_mode.get_conversion_rates(market_pair)
        if quote_rate is None:
            self.logger().warning(f"Can't find a conversion rate for {quote_pair}")
        if base_rate is None:
            self.logger().warning(f"Can't find a conversion rate for {base_pair}")
        if gas_rate is None:
            self.logger().warning(f"Can't find a conversion rate for {gas_pair}")
        return quote_pair, quote_rate_source, quote_rate, base_pair, base_rate_source, base_rate, gas_pair, \
            gas_rate_source, gas_rate

    def log_conversion_rates(self):
        for market_pair in self._market_pairs.values():
            quote_pair, quote_rate_source, quote_rate, base_pair, base_rate_source, base_rate, gas_pair, \
                gas_rate_source, gas_rate = self.get_conversion_rates(market_pair)
            if quote_pair.split("-")[0] != quote_pair.split("-")[1]:
                self.logger().info(f"{quote_pair} ({quote_rate_source}) conversion rate: {PerformanceMetrics.smart_round(quote_rate)}")
            if base_pair.split("-")[0] != base_pair.split("-")[1]:
                self.logger().info(f"{base_pair} ({base_rate_source}) conversion rate: {PerformanceMetrics.smart_round(base_rate)}")
            if self.is_gateway_market(market_pair.taker):
                if gas_pair is not None and gas_pair.split("-")[0] != gas_pair.split("-")[1]:
                    self.logger().info(f"{gas_pair} ({gas_rate_source}) conversion rate: {PerformanceMetrics.smart_round(gas_rate)}")

    def oracle_status_df(self):
        columns = ["Source", "Pair", "Rate"]
        data = []
        for market_pair in self._market_pairs.values():
            quote_pair, quote_rate_source, quote_rate, base_pair, base_rate_source, base_rate, gas_pair, \
                gas_rate_source, gas_rate = self.get_conversion_rates(market_pair)
            if quote_pair.split("-")[0] != quote_pair.split("-")[1]:
                data.extend([
                    [quote_rate_source, quote_pair, PerformanceMetrics.smart_round(quote_rate)],
                ])
            if base_pair.split("-")[0] != base_pair.split("-")[1]:
                data.extend([
                    [base_rate_source, base_pair, PerformanceMetrics.smart_round(base_rate)],
                ])
            if self.is_gateway_market(market_pair.taker):
                if gas_pair is not None and gas_pair.split("-")[0] != gas_pair.split("-")[1]:
                    data.extend([
                        [gas_rate_source, gas_pair, PerformanceMetrics.smart_round(gas_rate)],
                    ])
        return pd.DataFrame(data=data, columns=columns)

    def format_status(self) -> str:
        lines = []
        warning_lines = []
        tracked_maker_orders = {}

        # Go through the currently open limit orders, and group them by market pair.
        for market, limit_order, order_id in self.active_maker_limit_orders:
            typed_limit_order = limit_order
            market_pair = self._market_pair_tracker.get_market_pair_from_order_id(typed_limit_order.client_order_id)
            if market_pair not in tracked_maker_orders:
                tracked_maker_orders[market_pair] = {typed_limit_order.client_order_id: typed_limit_order}
            else:
                tracked_maker_orders[market_pair][typed_limit_order.client_order_id] = typed_limit_order

        for market_pair in self._market_pairs.values():
            warning_lines.extend(self.network_warning([market_pair.maker, market_pair.taker]))

            if not self.is_gateway_market(market_pair.taker):
                markets_df = self.market_status_data_frame([market_pair.maker, market_pair.taker])
            else:
                markets_df = self.market_status_data_frame([market_pair.maker])
                # Market status for gateway
                bid_price = "" if self._last_taker_buy_price is None else self._last_taker_buy_price
                ask_price = "" if self._last_taker_sell_price is None else self._last_taker_sell_price
                if self._last_taker_buy_price is not None and self._last_taker_sell_price is not None:
                    mid_price = (self._last_taker_buy_price + self._last_taker_sell_price) / 2
                else:
                    mid_price = ""
                taker_data = {
                    "Exchange": market_pair.taker.market.display_name,
                    "Market": market_pair.taker.trading_pair,
                    "Best Bid Price": bid_price,
                    "Best Ask Price": ask_price,
                    "Mid Price": mid_price
                }
                if markets_df is not None:
                    markets_df = pd.concat([markets_df, pd.DataFrame([taker_data])], ignore_index=True)
            lines.extend(["", "  Markets:"] +
                         ["    " + line for line in str(markets_df).split("\n")])

            oracle_df = self.oracle_status_df()
            if not oracle_df.empty:
                lines.extend(["", "  Rate conversion:"] +
                             ["    " + line for line in str(oracle_df).split("\n")])

            assets_df = self.wallet_balance_data_frame([market_pair.maker, market_pair.taker])
            lines.extend(["", "  Assets:"] +
                         ["    " + line for line in str(assets_df).split("\n")])

            # See if there're any open orders.
            if market_pair in tracked_maker_orders and len(tracked_maker_orders[market_pair]) > 0:
                limit_orders = list(tracked_maker_orders[market_pair].values())
                bid, ask = self.get_top_bid_ask(market_pair)
                mid_price = (bid + ask) / 2
                df = LimitOrder.to_pandas(limit_orders, float(mid_price))
                df_lines = str(df).split("\n")
                lines.extend(["", "  Active maker market orders:"] +
                             ["    " + line for line in df_lines])
            else:
                lines.extend(["", "  No active maker market orders."])

            warning_lines.extend(self.balance_warning([market_pair.maker, market_pair.taker]))

        if len(warning_lines) > 0:
            lines.extend(["", "  *** WARNINGS ***"] + warning_lines)

        return "\n".join(lines)

    def start(self, clock: Clock, timestamp: float):
        super().start(clock, timestamp)
        self._last_timestamp = timestamp

    def tick(self, timestamp: float):
        """
        Clock tick entry point.

        For cross exchange market making strategy, this function mostly just checks the readiness and connection
        status of markets, and then delegates the processing of each market pair to process_market_pair().

        :param timestamp: current tick timestamp
        """
        current_tick = (timestamp // self._status_report_interval)
        last_tick = (self._last_timestamp // self._status_report_interval)
        should_report_warnings = ((current_tick > last_tick) and
                                  (LogOption.STATUS_REPORT in self.logging_options)
                                  )

        # Perform clock tick with the market pair tracker.
        self._market_pair_tracker.tick(timestamp)

        if not self._all_markets_ready:
            self._all_markets_ready = all([market.ready for market in self.active_markets])
            if not self._all_markets_ready:
                # Markets not ready yet. Don't do anything.
                if should_report_warnings:
                    self.logger().warning("Markets are not ready. No market making trades are permitted.")
                return
            else:
                # Markets are ready, ok to proceed.
                if LogOption.STATUS_REPORT:
                    self.logger().info("Markets are ready.")

        if not self._conversions_ready:
            for market_pair in self._market_pairs.values():
                _, _, quote_rate, _, _, base_rate, _, _, _ = self.get_conversion_rates(market_pair)
                if not quote_rate or not base_rate:
                    if should_report_warnings:
                        self.logger().warning("Conversion rates are not ready. No market making trades are permitted.")
                    return

            # Conversion rates are ready, ok to proceed.
            self._conversions_ready = True
            if LogOption.STATUS_REPORT:
                self.logger().info("Conversion rates are ready. Trading started.")

        if should_report_warnings:
            # Check if all markets are still connected or not. If not, log a warning.
            if not all([market.network_status is NetworkStatus.CONNECTED for market in self.active_markets]):
                self.logger().warning("WARNING: Some markets are not connected or are down at the moment. Market "
                                      "making may be dangerous when markets or networks are unstable.")

        if self._gateway_quotes_task is None or self._gateway_quotes_task.done():
            self._gateway_quotes_task = safe_ensure_future(self.get_gateway_quotes())

        if self.ready_for_new_trades():
            if self._main_task is None or self._main_task.done():
                self._main_task = safe_ensure_future(self.main(timestamp))

    async def main(self, timestamp: float):
        try:
            # Calculate a mapping from market pair to list of active limit orders on the market.
            market_pair_to_active_orders = defaultdict(list)

            for maker_market, limit_order, order_id in self.active_maker_limit_orders:
                market_pair = self._market_pairs.get((maker_market, limit_order.trading_pair))
                if market_pair is None:
                    self.log_with_clock(logging.WARNING,
                                        f"The in-flight maker order in for the trading pair '{limit_order.trading_pair}' "
                                        f"does not correspond to any whitelisted trading pairs. Skipping.")
                    continue

                if not self._sb_order_tracker.has_in_flight_cancel(limit_order.client_order_id) and \
                        limit_order.client_order_id in self._maker_to_taker_order_ids.keys():
                    market_pair_to_active_orders[market_pair].append(limit_order)

            # Process each market pair independently.
            for market_pair in self._market_pairs.values():
                await self.process_market_pair(timestamp, market_pair, market_pair_to_active_orders[market_pair])

            # log conversion rates every 5 minutes
            if self._last_conv_rates_logged + (60. * 5) < timestamp:
                self.log_conversion_rates()
                self._last_conv_rates_logged = timestamp
        finally:
            self._last_timestamp = timestamp

    async def get_gateway_quotes(self):
        for market_pair in self._market_pairs.values():
            if self.is_gateway_market(market_pair.taker):
                _, _, quote_rate, _, _, base_rate, _, _, _ = self.get_conversion_rates(market_pair)
                order_amount = self._config_map.order_amount * base_rate
                order_price = await market_pair.taker.market.get_order_price(
                    market_pair.taker.trading_pair,
                    True,
                    order_amount
                )
                self._last_taker_buy_price = order_price
                order_price = await market_pair.taker.market.get_order_price(
                    market_pair.taker.trading_pair,
                    False,
                    order_amount
                )
                self._last_taker_sell_price = order_price

    def ready_for_new_trades(self) -> bool:
        """
        Returns True if there is no outstanding unfilled order.
        """
        if len(self._ongoing_hedging.keys()) > 0:
            return False
        return True

    def has_active_taker_order(self, market_pair: MarketTradingPairTuple):
        # Market orders are not being submitted as taker orders, limit orders are preferred at all times
        limit_orders = self._sb_order_tracker.get_limit_orders()
        limit_orders = limit_orders.get(market_pair, {})
        if len(limit_orders) > 0:
            if len(set(limit_orders.keys()).intersection(set(self._taker_to_maker_order_ids.keys()))) > 0:
                return True
        return False

    async def process_market_pair(self, timestamp: float, market_pair: MarketTradingPairTuple, active_orders: List):
        """
        For market pair being managed by this strategy object, do the following:

         1. Check whether any of the existing orders need to be canceled.
         2. Check if new orders should be created.

        For each market pair, only 1 active bid offer and 1 active ask offer is allowed at a time at maximum.

        If an active order is determined to be not needed at step 1, it would cancel the order within step 1.

        If there's no active order found in step 1, and condition allows (i.e. profitability, account balance, etc.),
        then a new limit order would be created at step 2.

        Combining step 1 and step 2 over time, means the offers made to the maker market side would be adjusted over
        time regularly.

        :param market_pair: cross exchange market pair
        :param active_orders: list of active maker limit orders associated with the market pair
        """
        has_active_bid = False
        has_active_ask = False
        need_adjust_order = False
        anti_hysteresis_timer = self._anti_hysteresis_timers.get(market_pair, 0)

        self.take_suggested_price_sample(timestamp, market_pair)

        for active_order in active_orders:
            # Mark the has_active_bid and has_active_ask flags
            is_buy = active_order.is_buy
            if is_buy:
                has_active_bid = True
            else:
                has_active_ask = True

            # Suppose the active order is hedged on the taker market right now, what's the average price the hedge
            # would happen?
            current_hedging_price = await self.calculate_effective_hedging_price(
                market_pair,
                is_buy,
                active_order.quantity
            )

            # See if it's still profitable to keep the order on maker market. If not, remove it.
            if not await self.check_if_still_profitable(market_pair, active_order, current_hedging_price):
                continue

            if isinstance(self._config_map.order_refresh_mode, PassiveOrderRefreshMode):
                continue

            # See if I still have enough balance on my wallet to fill the order on maker market, and to hedge the
            # order on taker market. If not, adjust it.
            if not await self.check_if_sufficient_balance(market_pair, active_order):
                continue

            # If prices have moved, one side is still profitable, here cancel and
            # place at the next tick.
            if timestamp > anti_hysteresis_timer:
                if not await self.check_if_price_has_drifted(market_pair, active_order):
                    need_adjust_order = True
                    continue

        # If order adjustment is needed in the next tick, set the anti-hysteresis timer s.t. the next order adjustment
        # for the same pair wouldn't happen within the time limit.
        if need_adjust_order:
            self._anti_hysteresis_timers[market_pair] = timestamp + self._config_map.anti_hysteresis_duration

        # If there's both an active bid and ask, then there's no need to think about making new limit orders.
        if has_active_bid and has_active_ask:
            return

        # If there are pending taker orders, wait for them to complete
        if self.has_active_taker_order(market_pair):
            return

        # See if it's profitable to place a limit order on maker market.
        await self.check_and_create_new_orders(market_pair, has_active_bid, has_active_ask)

    async def hedge_filled_maker_order(self,
                                       maker_order_id: str,
                                       order_filled_event: OrderFilledEvent):
        """
        If a limit order previously made to the maker side has been filled, hedge it on the taker side.
        :param order_filled_event: event object
        """
        order_id = order_filled_event.order_id
        market_pair = self._market_pair_tracker.get_market_pair_from_order_id(order_id)

        # Make sure to only hedge limit orders.
        if market_pair is not None and order_id not in self._taker_to_maker_order_ids.keys():
            limit_order_record = self._sb_order_tracker.get_shadow_limit_order(order_id)
            order_fill_record = (limit_order_record, order_filled_event)

            # Store the limit order fill event in a map, s.t. it can be processed in check_and_hedge_orders()
            # later.
            if order_filled_event.trade_type is TradeType.BUY:
                if market_pair not in self._order_fill_buy_events:
                    self._order_fill_buy_events[market_pair] = [order_fill_record]
                else:
                    self._order_fill_buy_events[market_pair].append(order_fill_record)

                if LogOption.MAKER_ORDER_FILLED in self.logging_options:
                    self.log_with_clock(
                        logging.INFO,
                        f"({market_pair.maker.trading_pair}) Maker buy order of "
                        f"{order_filled_event.amount} {market_pair.maker.base_asset} filled."
                    )

            else:
                if market_pair not in self._order_fill_sell_events:
                    self._order_fill_sell_events[market_pair] = [order_fill_record]
                else:
                    self._order_fill_sell_events[market_pair].append(order_fill_record)

                if LogOption.MAKER_ORDER_FILLED in self.logging_options:
                    self.log_with_clock(
                        logging.INFO,
                        f"({market_pair.maker.trading_pair}) Maker sell order of "
                        f"{order_filled_event.amount} {market_pair.maker.base_asset} filled."
                    )

            # Call check_and_hedge_orders() to emit the orders on the taker side.
            try:
                await self.check_and_hedge_orders(maker_order_id, market_pair)
            except Exception:
                self.log_with_clock(logging.ERROR, "Unexpected error.", exc_info=True)

    def hedge_tasks_cleanup(self):
        hedge_maker_order_tasks = []
        for task in self._hedge_maker_order_tasks:
            if not task.done():
                hedge_maker_order_tasks += [task]
        self._hedge_maker_order_tasks = hedge_maker_order_tasks

    def handle_unfilled_taker_order(self, order_event):
        order_id = order_event.order_id
        market_pair = self._market_pair_tracker.get_market_pair_from_order_id(order_id)

        # Resubmit hedging order
        self.hedge_tasks_cleanup()
        self._hedge_maker_order_tasks += [safe_ensure_future(
            self.check_and_hedge_orders(order_id, market_pair)
        )]

        # Remove the cancelled, failed or expired taker order
        del self._taker_to_maker_order_ids[order_event.order_id]

    def did_fill_order(self, order_filled_event: OrderFilledEvent):
        maker_order_id = order_filled_event.order_id
        exchange_trade_id = order_filled_event.exchange_trade_id
        if maker_order_id in self._maker_to_taker_order_ids.keys():
            # Maker order filled
            # Check if this fill was already processed or not
            if maker_order_id not in self._maker_to_hedging_trades.keys():
                self._maker_to_hedging_trades[maker_order_id] = []
            if exchange_trade_id not in self._maker_to_hedging_trades[maker_order_id]:
                # This maker fill has not been processed yet, submit Taker hedge order
                # Values have to be unique in a bidict

                self._maker_to_hedging_trades[maker_order_id] += [exchange_trade_id]

                self.hedge_tasks_cleanup()
                self._hedge_maker_order_task = safe_ensure_future(
                    self.hedge_filled_maker_order(maker_order_id, order_filled_event)
                )

    def did_cancel_order(self, order_canceled_event: OrderCancelledEvent):
        if order_canceled_event.order_id in self._taker_to_maker_order_ids.keys():
            self.handle_unfilled_taker_order(order_canceled_event)

    def did_fail_order(self, order_failed_event: MarketOrderFailureEvent):
        if order_failed_event.order_id in self._taker_to_maker_order_ids.keys():
            self.handle_unfilled_taker_order(order_failed_event)

    def did_expire_order(self, order_expired_event: OrderExpiredEvent):
        if order_expired_event.order_id in self._taker_to_maker_order_ids.keys():
            self.handle_unfilled_taker_order(order_expired_event)

    def did_complete_buy_order(self, order_completed_event: BuyOrderCompletedEvent):
        """
        Output log message when a bid order (on maker side or taker side) is completely taken.
        :param order_completed_event: event object
        """
        order_id = order_completed_event.order_id
        market_pair = self._market_pair_tracker.get_market_pair_from_order_id(order_id)

        if market_pair is not None:
            if order_id in self._maker_to_taker_order_ids.keys():
                limit_order_record = self._sb_order_tracker.get_limit_order(market_pair.maker, order_id)
                self.log_with_clock(
                    logging.INFO,
                    f"({market_pair.maker.trading_pair}) Maker buy order {order_id} "
                    f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
                    f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
                )
                self.notify_hb_app_with_timestamp(
                    f"Maker BUY order ({limit_order_record.quantity} {limit_order_record.base_currency} @ "
                    f"{limit_order_record.price} {limit_order_record.quote_currency}) is filled."
                )
                # Leftover other side maker order will be left in the market until its expiration or potential fill
                # Since the buy side side was filled, the sell side maker order is unlikely to be filled, therefore
                # it'll likey expire
                # The others are left in the market to collect market making fees
                # Meanwhile new maker order may be placed
            if order_id in self._taker_to_maker_order_ids.keys():
                self.log_with_clock(
                    logging.INFO,
                    f"({market_pair.taker.trading_pair}) Taker buy order {order_id} for "
                    f"({order_completed_event.base_asset_amount} {order_completed_event.base_asset} has been completely filled."
                )
                self.notify_hb_app_with_timestamp(
                    f"Taker BUY order ({order_completed_event.base_asset_amount} {order_completed_event.base_asset} "
                    f"{order_completed_event.quote_asset}) is filled."
                )
                maker_order_id = self._taker_to_maker_order_ids[order_id]
                # Remove the completed taker order
                del self._taker_to_maker_order_ids[order_id]
                # Get all active taker order ids for the maker order id
                active_taker_ids = set(self._taker_to_maker_order_ids.keys()).intersection(set(
                    self._maker_to_taker_order_ids[maker_order_id]))
                if len(active_taker_ids) == 0:
                    # Was maker order fully filled?
                    maker_order_ids = list(order_id for market, limit_order, order_id in self.active_maker_limit_orders)
                    if maker_order_id not in maker_order_ids:
                        # Remove the completed fully hedged maker order
                        del self._maker_to_taker_order_ids[maker_order_id]
                        del self._maker_to_hedging_trades[maker_order_id]

                try:
                    self.del_order_from_ongoing_hedging(order_id)
                except KeyError:
                    self.logger().warning(f"Ongoing hedging not found for order id {order_id}")

                # Delete hedged maker fill event
                fill_events = []
                for fill_event in self._order_fill_sell_events[market_pair]:
                    if self.is_fill_event_in_ongoing_hedging(fill_event):
                        fill_events += [fill_event]
                self._order_fill_sell_events[market_pair] = fill_events

                # Cleanup maker fill events - no longer needed to create taker orders if all fills were hedged
                if len(self._order_fill_sell_events[market_pair]) == 0:
                    del self._order_fill_sell_events[market_pair]

    def did_complete_sell_order(self, order_completed_event: SellOrderCompletedEvent):
        """
        Output log message when a ask order (on maker side or taker side) is completely taken.
        :param order_completed_event: event object
        """
        order_id = order_completed_event.order_id
        market_pair = self._market_pair_tracker.get_market_pair_from_order_id(order_id)

        if market_pair is not None:
            if order_id in self._maker_to_taker_order_ids.keys():
                limit_order_record = self._sb_order_tracker.get_limit_order(market_pair.maker, order_id)
                self.log_with_clock(
                    logging.INFO,
                    f"({market_pair.maker.trading_pair}) Maker sell order {order_id} "
                    f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
                    f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
                )
                self.notify_hb_app_with_timestamp(
                    f"Maker sell order ({limit_order_record.quantity} {limit_order_record.base_currency} @ "
                    f"{limit_order_record.price} {limit_order_record.quote_currency}) is filled."
                )
                # Leftover other side maker order will be left in the market until its expiration or potential fill
                # Since the sell side side was filled, the buy side maker order is unlikely to be filled, therefore
                # it'll likey expire
                # The others are left in the market to collect market making fees
                # Meanwhile new maker order may be placed
            if order_id in self._taker_to_maker_order_ids.keys():
                self.log_with_clock(
                    logging.INFO,
                    f"({market_pair.taker.trading_pair}) Taker sell order {order_id} for "
                    f"({order_completed_event.base_asset_amount} {order_completed_event.base_asset} "
                    f"has been completely filled."
                )
                self.notify_hb_app_with_timestamp(
                    f"Taker SELL order ({order_completed_event.base_asset_amount} {order_completed_event.base_asset} "
                    f"{order_completed_event.quote_asset}) is filled."
                )
                maker_order_id = self._taker_to_maker_order_ids[order_id]
                # Remove the completed taker order
                del self._taker_to_maker_order_ids[order_id]
                # Get all active taker order ids for the maker order id
                active_taker_ids = set(self._taker_to_maker_order_ids.keys()).intersection(set(
                    self._maker_to_taker_order_ids[maker_order_id]))
                if len(active_taker_ids) == 0:
                    # Was maker order fully filled?
                    maker_order_ids = list(order_id for market, limit_order, order_id in self.active_maker_limit_orders)
                    if maker_order_id not in maker_order_ids:
                        # Remove the completed fully hedged maker order
                        del self._maker_to_taker_order_ids[maker_order_id]
                        del self._maker_to_hedging_trades[maker_order_id]

                try:
                    self.del_order_from_ongoing_hedging(order_id)
                except KeyError:
                    self.logger().warning(f"Ongoing hedging not found for order id {order_id}")

                # Delete hedged maker fill event
                fill_events = []
                for fill_event in self._order_fill_buy_events[market_pair]:
                    if self.is_fill_event_in_ongoing_hedging(fill_event):
                        fill_events += [fill_event]
                self._order_fill_buy_events[market_pair] = fill_events

                # Cleanup maker fill events - no longer needed to create taker orders if all fills were hedged
                if len(self._order_fill_buy_events[market_pair]) == 0:
                    del self._order_fill_buy_events[market_pair]

    async def check_if_price_has_drifted(self, market_pair: MakerTakerMarketPair, active_order: LimitOrder):
        """
        Given a currently active limit order on maker side, check if its current price is still valid, based on the
        current hedging price on taker market, depth tolerance, and transient orders on the maker market captured by
        recent suggested price samples.

        If the active order's price is no longer valid, the order will be canceled.

        This function is only used when active order cancelation is enabled.

        :param market_pair: cross exchange market pair
        :param active_order: a current active limit order in the market pair
        :return: True if the order stays, False if the order has been canceled and we need to re place the orders.
        """
        is_buy = active_order.is_buy
        order_price = active_order.price
        order_quantity = active_order.quantity
        suggested_price = await self.get_market_making_price(market_pair, is_buy, order_quantity)

        if suggested_price != order_price:
            if LogOption.ADJUST_ORDER in self.logging_options:
                self.log_with_clock(
                    logging.INFO,
                    f"({market_pair.maker.trading_pair}) The current limit {'bid' if is_buy else 'ask'} order for "
                    f"{active_order.quantity} {market_pair.maker.base_asset} at "
                    f"{order_price:.8g} {market_pair.maker.quote_asset} is now below the suggested order "
                    f"price at {suggested_price}. Going to cancel the old order and create a new one..."
                )
            self.cancel_maker_order(market_pair, active_order.client_order_id)
            self.log_with_clock(logging.DEBUG,
                                f"Current {'buy' if is_buy else 'sell'} order price={order_price}, "
                                f"suggested order price={suggested_price}")
            return False

        return True

    async def check_and_hedge_orders(self,
                                     maker_order_id: str,
                                     market_pair: MakerTakerMarketPair):
        """
        Look into the stored and un-hedged limit order fill events, and emit orders to hedge them, depending on
        availability of funds on the taker market.

        :param market_pair: cross exchange market pair
        """

        buy_fill_records = self.get_unhedged_buy_records(market_pair)
        sell_fill_records = self.get_unhedged_sell_records(market_pair)

        buy_fill_quantity = sum([fill_event.amount for _, fill_event in buy_fill_records])
        sell_fill_quantity = sum([fill_event.amount for _, fill_event in sell_fill_records])

        taker_trading_pair = market_pair.taker.trading_pair
        taker_market = market_pair.taker.market

        # Convert maker order size (in maker base asset) to taker order size (in taker base asset)
        _, _, quote_rate, _, _, base_rate, _, _, _ = self.get_conversion_rates(market_pair)

        if buy_fill_quantity > 0:
            # Maker buy
            # Taker sell
            taker_slippage_adjustment_factor = Decimal("1") - self.slippage_buffer

            hedged_order_quantity = min(
                buy_fill_quantity / base_rate,
                taker_market.get_available_balance(market_pair.taker.base_asset) *
                self.order_size_taker_balance_factor
            )
            quantized_hedge_amount = taker_market.quantize_order_amount(taker_trading_pair, Decimal(hedged_order_quantity))

            avg_fill_price = (sum([r.price * r.amount for _, r in buy_fill_records]) /
                              sum([r.amount for _, r in buy_fill_records]))

            self.check_multiple_buy_orders(buy_fill_records)
            if self.is_gateway_market(market_pair.taker):
                order_price = await market_pair.taker.market.get_order_price(
                    taker_trading_pair,
                    False,
                    quantized_hedge_amount)
                if order_price is None:
                    self.logger().warning("Gateway: failed to obtain order price. No hedging order will be submitted.")
                    return
                taker_top = order_price
            else:
                taker_top = taker_market.get_price(taker_trading_pair, False)
                order_price = taker_market.get_price_for_volume(
                    taker_trading_pair, False, quantized_hedge_amount
                ).result_price

            self.log_with_clock(logging.INFO, f"Calculated by HB order_price: {order_price}")
            order_price *= taker_slippage_adjustment_factor
            order_price = taker_market.quantize_order_price(taker_trading_pair, order_price)
            self.log_with_clock(logging.INFO, f"Slippage buffer adjusted order_price: {order_price}")

            if quantized_hedge_amount > s_decimal_zero:
                self.place_order(
                    market_pair,
                    False,
                    False,
                    quantized_hedge_amount,
                    order_price,
                    maker_order_id,
                    buy_fill_records
                )

                if LogOption.MAKER_ORDER_HEDGED in self.logging_options:
                    self.log_with_clock(
                        logging.INFO,
                        f"({market_pair.maker.trading_pair}) Hedged maker buy order(s) of "
                        f"{buy_fill_quantity} {market_pair.maker.base_asset} on taker market to lock in profits. "
                        f"(maker avg price={avg_fill_price}, taker top={taker_top})"
                    )
            else:
                self.log_with_clock(
                    logging.INFO,
                    f"({market_pair.maker.trading_pair}) Current maker buy fill amount of "
                    f"{buy_fill_quantity} {market_pair.maker.base_asset} is less than the minimum order amount "
                    f"allowed on the taker market. No hedging possible yet."
                )

        if sell_fill_quantity > 0:
            # Maker sell
            # Taker buy
            taker_slippage_adjustment_factor = Decimal("1") + self.slippage_buffer

            if self.is_gateway_market(market_pair.taker):
                taker_price = await market_pair.taker.market.get_order_price(
                    taker_trading_pair,
                    True,
                    sell_fill_quantity / base_rate
                )
                if taker_price is None:
                    self.logger().warning("Gateway: failed to obtain order price. No hedging order will be submitted.")
                    return
            else:
                taker_price = taker_market.get_price_for_volume(
                    taker_trading_pair,
                    True,
                    sell_fill_quantity / base_rate
                ).result_price

            hedged_order_quantity = min(
                sell_fill_quantity / base_rate,
                taker_market.get_available_balance(market_pair.taker.quote_asset) /
                taker_price * self.order_size_taker_balance_factor
            )
            quantized_hedge_amount = taker_market.quantize_order_amount(
                taker_trading_pair,
                Decimal(hedged_order_quantity)
            )

            avg_fill_price = (sum([r.price * r.amount for _, r in sell_fill_records]) /
                              sum([r.amount for _, r in sell_fill_records]))

            self.check_multiple_sell_orders(sell_fill_records)
            if self.is_gateway_market(market_pair.taker):
                order_price = await market_pair.taker.market.get_order_price(
                    taker_trading_pair,
                    True,
                    quantized_hedge_amount)
                if order_price is None:
                    self.logger().warning("Gateway: failed to obtain order price. No hedging order will be submitted.")
                    return
                taker_top = order_price
            else:
                taker_top = taker_market.get_price(taker_trading_pair, True)
                order_price = taker_market.get_price_for_volume(
                    taker_trading_pair, True, quantized_hedge_amount
                ).result_price

            self.log_with_clock(logging.INFO, f"Calculated by HB order_price: {order_price}")
            order_price *= taker_slippage_adjustment_factor
            order_price = taker_market.quantize_order_price(taker_trading_pair, order_price)
            self.log_with_clock(logging.INFO, f"Slippage buffer adjusted order_price: {order_price}")

            if quantized_hedge_amount > s_decimal_zero:
                self.place_order(
                    market_pair,
                    True,
                    False,
                    quantized_hedge_amount,
                    order_price,
                    maker_order_id,
                    sell_fill_records,
                )

                if LogOption.MAKER_ORDER_HEDGED in self.logging_options:
                    self.log_with_clock(
                        logging.INFO,
                        f"({market_pair.maker.trading_pair}) Hedged maker sell order(s) of "
                        f"{sell_fill_quantity} {market_pair.maker.base_asset} on taker market to lock in profits. "
                        f"(maker avg price={avg_fill_price}, taker top={taker_top})"
                    )
            else:
                self.log_with_clock(
                    logging.INFO,
                    f"({market_pair.maker.trading_pair}) Current maker sell fill amount of "
                    f"{sell_fill_quantity} {market_pair.maker.base_asset} is less than the minimum order amount "
                    f"allowed on the taker market. No hedging possible yet."
                )

    def get_adjusted_limit_order_size(self, market_pair: MakerTakerMarketPair) -> Tuple[Decimal, Decimal]:
        """
        Given the proposed order size of a proposed limit order (regardless of bid or ask), adjust and refine the order
        sizing according to either the trade size override setting (if it exists), or the portfolio ratio limit (if
        no trade size override exists).

        Also, this function will convert the input order size proposal from floating point to Decimal by quantizing the
        order size.

        :param market_pair: cross exchange market pair
        :rtype: Decimal
        """
        maker_market = market_pair.maker.market
        trading_pair = market_pair.maker.trading_pair
        if self._config_map.order_amount and self._config_map.order_amount > 0:
            base_order_size = self._config_map.order_amount
            return maker_market.quantize_order_amount(trading_pair, Decimal(base_order_size))
        else:
            return self.get_order_size_after_portfolio_ratio_limit(market_pair)

    def get_order_size_after_portfolio_ratio_limit(self, market_pair: MakerTakerMarketPair) -> Decimal:
        """
        Given the proposed order size of a proposed limit order (regardless of bid or ask), adjust the order sizing
        according to the portfolio ratio limit.

        Also, this function will convert the input order size proposal from floating point to Decimal by quantizing the
        order size.

        :param market_pair: cross exchange market pair
        :rtype: Decimal
        """
        maker_market = market_pair.maker.market
        trading_pair = market_pair.maker.trading_pair
        base_balance = maker_market.get_balance(market_pair.maker.base_asset)
        quote_balance = maker_market.get_balance(market_pair.maker.quote_asset)
        current_price = (maker_market.get_price(trading_pair, True) +
                         maker_market.get_price(trading_pair, False)) * Decimal(0.5)
        maker_portfolio_value = base_balance + quote_balance / current_price
        adjusted_order_size = maker_portfolio_value * self.order_size_portfolio_ratio_limit

        return maker_market.quantize_order_amount(trading_pair, Decimal(adjusted_order_size))

    async def get_market_making_size(self,
                                     market_pair: MakerTakerMarketPair,
                                     is_bid: bool):
        """
        Get the ideal market making order size given a market pair and a side.

        This function does a few things:
         1. Calculate the largest order size possible given the current balances on both maker and taker markets.
         2. Calculate the largest order size possible that's still profitable after hedging.


        :param market_pair: The cross exchange market pair to calculate order price/size limits.
        :param is_bid: Whether the order to make will be bid or ask.
        :return: a Decimal which is the size of maker order.
        """
        taker_trading_pair = market_pair.taker.trading_pair
        maker_market = market_pair.maker.market
        taker_market = market_pair.taker.market

        # Maker order size (in base asset)
        size = self.get_adjusted_limit_order_size(market_pair)

        # Convert maker order size (in maker base asset) to taker order size (in taker base asset)
        _, _, quote_rate, _, _, base_rate, _, _, _ = self.get_conversion_rates(market_pair)
        taker_size = size / base_rate

        if is_bid:
            # Maker buy
            # Taker sell
            maker_balance_in_quote = maker_market.get_available_balance(market_pair.maker.quote_asset)
            taker_balance = taker_market.get_available_balance(market_pair.taker.base_asset) * \
                self.order_size_taker_balance_factor

            if self.is_gateway_market(market_pair.taker):
                taker_price = await taker_market.get_order_price(taker_trading_pair,
                                                                 False,
                                                                 taker_size)
                if taker_price is None:
                    self.logger().warning("Gateway: failed to obtain order price."
                                          "No market making order will be submitted.")
                    return s_decimal_zero
            else:
                try:
                    taker_price = taker_market.get_vwap_for_volume(
                        taker_trading_pair, False, taker_size
                    ).result_price
                except ZeroDivisionError:
                    assert size == s_decimal_zero
                    return s_decimal_zero

            if taker_price is None:
                self.logger().warning("Failed to obtain a taker sell order price. No order will be submitted.")
                order_amount = Decimal("0")
            else:
                maker_balance = maker_balance_in_quote / \
                    (taker_price * self.markettaker_to_maker_base_conversion_rate(market_pair))
                taker_balance *= base_rate
                order_amount = min(maker_balance, taker_balance, size)

            return maker_market.quantize_order_amount(market_pair.maker.trading_pair, Decimal(order_amount))

        else:
            # Maker sell
            # Taker buy
            maker_balance = maker_market.get_available_balance(market_pair.maker.base_asset)
            taker_balance_in_quote = taker_market.get_available_balance(market_pair.taker.quote_asset) * \
                self.order_size_taker_balance_factor

            if self.is_gateway_market(market_pair.taker):
                taker_price = await taker_market.get_order_price(taker_trading_pair,
                                                                 True,
                                                                 size)
                if taker_price is None:
                    self.logger().warning("Gateway: failed to obtain order price."
                                          "No market making order will be submitted.")
                    return s_decimal_zero
            else:
                try:
                    taker_price = taker_market.get_price_for_quote_volume(
                        taker_trading_pair, True, taker_balance_in_quote
                    ).result_price
                except ZeroDivisionError:
                    assert size == s_decimal_zero
                    return s_decimal_zero

            if taker_price is None:
                self.logger().warning("Failed to obtain a taker buy order price. No order will be submitted.")
                order_amount = Decimal("0")
            else:
                taker_slippage_adjustment_factor = Decimal("1") + self.slippage_buffer
                taker_balance = taker_balance_in_quote / (taker_price * taker_slippage_adjustment_factor)
                taker_balance *= base_rate
                order_amount = min(maker_balance, taker_balance, size)

            return maker_market.quantize_order_amount(market_pair.maker.trading_pair, Decimal(order_amount))

    async def get_market_making_price(self,
                                      market_pair: MarketTradingPairTuple,
                                      is_bid: bool,
                                      size: Decimal):
        """
        Get the ideal market making order price given a market pair, side and size.

        The price returned is calculated by adding the profitability to vwap of hedging it on the taker side.
        or if it's not possible to hedge the trade profitably, then the returned order price will be none.

        :param market_pair: The cross exchange market pair to calculate order price/size limits.
        :param is_bid: Whether the order to make will be bid or ask.
        :param size: size of the maker order.
        :return: a Decimal which is the price or None if order cannot be hedged on the taker market
        """
        taker_trading_pair = market_pair.taker.trading_pair
        maker_market = market_pair.maker.market
        taker_market = market_pair.taker.market
        top_bid_price = s_decimal_nan
        top_ask_price = s_decimal_nan
        next_price_below_top_ask = s_decimal_nan

        # Convert maker order size (in maker base asset) to taker order size (in taker base asset)
        _, _, quote_rate, base_pair, _, base_rate, _, _, _ = self.get_conversion_rates(market_pair)
        size /= base_rate

        top_bid_price, top_ask_price = self.get_top_bid_ask_from_price_samples(market_pair)

        if is_bid:
            # Maker buy
            # Taker sell
            if not Decimal.is_nan(top_bid_price):
                # Calculate the next price above top bid
                price_quantum = maker_market.get_order_price_quantum(
                    market_pair.maker.trading_pair,
                    top_bid_price
                )
                price_above_bid = (ceil(top_bid_price / price_quantum) + 1) * price_quantum

            if self.is_gateway_market(market_pair.taker):
                taker_price = await taker_market.get_order_price(taker_trading_pair,
                                                                 False,
                                                                 size)
                if taker_price is None:
                    self.logger().warning("Gateway: failed to obtain order price."
                                          "No market making order will be submitted.")
                    return s_decimal_nan
            else:
                try:
                    taker_price = taker_market.get_vwap_for_volume(taker_trading_pair, False, size).result_price
                except ZeroDivisionError:
                    return s_decimal_nan

            if taker_price is None:
                return

            # Convert them from taker's price to maker's price
            taker_price *= self.markettaker_to_maker_base_conversion_rate(market_pair)

            # you are buying on the maker market and selling on the taker market
            maker_price = taker_price / (1 + self.min_profitability)

            # # If your bid is higher than highest bid price, reduce it to one tick above the top bid price
            if self.adjust_order_enabled:
                # If maker bid order book is not empty
                if not Decimal.is_nan(price_above_bid):
                    maker_price = min(maker_price, price_above_bid)

            price_quantum = maker_market.get_order_price_quantum(
                market_pair.maker.trading_pair,
                maker_price
            )

            # Rounds down for ensuring profitability
            maker_price = (floor(maker_price / price_quantum)) * price_quantum

            return maker_price
        else:
            # Maker sell
            # Taker buy
            if not Decimal.is_nan(top_ask_price):
                # Calculate the next price below top ask
                price_quantum = maker_market.get_order_price_quantum(
                    market_pair.maker.trading_pair,
                    top_ask_price
                )
                next_price_below_top_ask = (floor(top_ask_price / price_quantum) - 1) * price_quantum

            if self.is_gateway_market(market_pair.taker):
                taker_price = await taker_market.get_order_price(taker_trading_pair,
                                                                 True,
                                                                 size)
                if taker_price is None:
                    self.logger().warning("Gateway: failed to obtain order price."
                                          "No market making order will be submitted.")
                    return s_decimal_nan
            else:
                try:
                    taker_price = taker_market.get_vwap_for_volume(taker_trading_pair, True, size).result_price
                except ZeroDivisionError:
                    return s_decimal_nan

            taker_price *= self.markettaker_to_maker_base_conversion_rate(market_pair)

            # You are selling on the maker market and buying on the taker market
            maker_price = taker_price * (1 + self.min_profitability)

            # If your ask is lower than the top ask, increase it to just one tick below top ask
            if self.adjust_order_enabled:
                # If maker ask order book is not empty
                if not Decimal.is_nan(next_price_below_top_ask):
                    maker_price = max(maker_price, next_price_below_top_ask)

            price_quantum = maker_market.get_order_price_quantum(
                market_pair.maker.trading_pair,
                maker_price
            )

            # Rounds up for ensuring profitability
            maker_price = (ceil(maker_price / price_quantum)) * price_quantum

            return maker_price

    async def calculate_effective_hedging_price(self,
                                                market_pair: MarketTradingPairTuple,
                                                is_bid: bool,
                                                size: Decimal):
        """
        Returns current possible taker price expressed in units of the maker market quote asset
        :param market_pair: The cross exchange market pair to calculate order price/size limits.
        :param is_bid: Whether the order to make will be bid or ask.
        :param size: The size of the maker order.
        :return: a Decimal which is the hedging price on the maker market given the current taker market
        """
        taker_trading_pair = market_pair.taker.trading_pair
        taker_market = market_pair.taker.market

        # Convert maker order size (in maker base asset) to taker order size (in taker base asset)
        _, _, quote_rate, _, _, base_rate, _, _, gas_rate = self.get_conversion_rates(market_pair)
        size /= base_rate

        # Calculate the next price from the top, and the order size limit.
        if is_bid:
            # Maker buy
            # Taker sell
            if self.is_gateway_market(market_pair.taker):
                taker_price = await taker_market.get_order_price(taker_trading_pair,
                                                                 False,
                                                                 size)
                if taker_price is None:
                    self.logger().warning("Gateway: failed to obtain order price."
                                          "Failed to calculate effective hedging price.")
                    return s_decimal_nan
            else:
                try:
                    taker_price = taker_market.get_vwap_for_volume(taker_trading_pair, False, size).result_price
                except ZeroDivisionError:
                    return None

            # If quote assets are not same, convert them from taker's quote asset to maker's quote asset
            taker_price *= self.markettaker_to_maker_base_conversion_rate(market_pair)

            return taker_price
        else:
            # Maker sell
            # Taker buy
            if self.is_gateway_market(market_pair.taker):
                taker_price = await taker_market.get_order_price(taker_trading_pair,
                                                                 True,
                                                                 size)
                if taker_price is None:
                    self.logger().warning("Gateway: failed to obtain order price."
                                          "Failed to calculate effective hedging price.")
                    return s_decimal_nan
            else:
                try:
                    taker_price = taker_market.get_vwap_for_volume(taker_trading_pair, True, size).result_price
                except ZeroDivisionError:
                    return None

            # Convert them from taker's price to maker's price
            taker_price *= self.markettaker_to_maker_base_conversion_rate(market_pair)

            return taker_price

    def get_suggested_price_samples(self, market_pair: MakerTakerMarketPair):
        """
        Get the queues of order book price samples for a market pair.

        :param market_pair: The market pair under which samples were collected for.
        :return: (bid order price samples, ask order price samples)
        """
        if market_pair in self._suggested_price_samples:
            return self._suggested_price_samples[market_pair]
        return deque(), deque()

    def get_top_bid_ask(self, market_pair: MakerTakerMarketPair):
        """
        Calculate the top bid and ask using top depth tolerance in maker order book

        :param market_pair: cross exchange market pair
        :return: (top bid: Decimal, top ask: Decimal)
        """
        trading_pair = market_pair.maker.trading_pair
        maker_market = market_pair.maker.market

        if self._config_map.top_depth_tolerance == 0:
            top_bid_price = maker_market.get_price(trading_pair, False)
            top_ask_price = maker_market.get_price(trading_pair, True)

        else:
            # Use bid entries in maker order book
            top_bid_price = maker_market.get_price_for_volume(trading_pair,
                                                              False,
                                                              self._config_map.top_depth_tolerance).result_price

            # Use ask entries in maker order book
            top_ask_price = maker_market.get_price_for_volume(trading_pair,
                                                              True,
                                                              self._config_map.top_depth_tolerance).result_price

        return top_bid_price, top_ask_price

    def take_suggested_price_sample(self, timestamp: float, market_pair: MakerTakerMarketPair):
        """
        Record the bid and ask sample queues.

        These samples are later taken to check if price has drifted for new limit orders, s.t. new limit orders can
        properly take into account transient orders that appear and disappear frequently on the maker market.

        :param market_pair: cross exchange market pair
        """
        if ((self._last_timestamp // self.ORDER_ADJUST_SAMPLE_INTERVAL) <
                (timestamp // self.ORDER_ADJUST_SAMPLE_INTERVAL)):
            if market_pair not in self._suggested_price_samples:
                self._suggested_price_samples[market_pair] = (deque(), deque())

            top_bid_price, top_ask_price = self.get_top_bid_ask_from_price_samples(market_pair)

            bid_price_samples_deque, ask_price_samples_deque = self._suggested_price_samples[market_pair]
            bid_price_samples_deque.append(top_bid_price)
            ask_price_samples_deque.append(top_ask_price)
            while len(bid_price_samples_deque) > self.ORDER_ADJUST_SAMPLE_WINDOW:
                bid_price_samples_deque.popleft()
            while len(ask_price_samples_deque) > self.ORDER_ADJUST_SAMPLE_WINDOW:
                ask_price_samples_deque.popleft()

    def get_top_bid_ask_from_price_samples(self, market_pair: MakerTakerMarketPair):
        """
        Calculate the top bid and ask using earlier samples

        :param market_pair: cross exchange market pair
        :return: (top bid, top ask)
        """
        # Incorporate the past bid & ask price samples.
        current_top_bid_price, current_top_ask_price = self.get_top_bid_ask(market_pair)

        bid_price_samples, ask_price_samples = self.get_suggested_price_samples(market_pair)

        if not any(Decimal.is_nan(p) for p in bid_price_samples) and not Decimal.is_nan(current_top_bid_price):
            top_bid_price = max(list(bid_price_samples) + [current_top_bid_price])
        else:
            top_bid_price = current_top_bid_price

        if not any(Decimal.is_nan(p) for p in ask_price_samples) and not Decimal.is_nan(current_top_ask_price):
            top_ask_price = min(list(ask_price_samples) + [current_top_ask_price])
        else:
            top_ask_price = current_top_ask_price

        return top_bid_price, top_ask_price

    async def check_if_still_profitable(self,
                                        market_pair,
                                        active_order: LimitOrder,
                                        current_hedging_price: Decimal):
        """
        Check whether a currently active limit order should be canceled or not, according to profitability metric.

        If active order canceling is enabled (e.g. for centralized exchanges), then the min profitability config is
        used as the threshold. If it is disabled (e.g. for decentralized exchanges), then the cancel order threshold
        is used instead.

        :param market_pair: cross exchange market pair
        :param active_order: the currently active order to check for cancelation
        :param current_hedging_price: the current average hedging price on taker market for the limit order
        :return: True if the limit order stays, False if the limit order is being canceled.
        """
        is_buy = active_order.is_buy
        limit_order_type_str = "bid" if is_buy else "ask"
        order_price = active_order.price

        cancel_order_threshold = self._config_map.order_refresh_mode.get_cancel_order_threshold()
        if cancel_order_threshold.is_nan():
            cancel_order_threshold = self.min_profitability

        if current_hedging_price is None:
            if LogOption.REMOVING_ORDER in self.logging_options:
                self.log_with_clock(
                    logging.INFO,
                    f"({market_pair.maker.trading_pair}) Limit {limit_order_type_str} order at "
                    f"{order_price:.8g} {market_pair.maker.quote_asset} is no longer profitable. "
                    f"Removing the order."
                )
            self.cancel_maker_order(market_pair, active_order.client_order_id)
            return False

        transaction_fee = 0
        pl = 0
        if self.is_gateway_market(market_pair.taker):
            if hasattr(market_pair.taker.market, "network_transaction_fee"):
                _, _, quote_rate, _, _, base_rate, _, _, gas_rate = self.get_conversion_rates(market_pair)
                transaction_fee: TokenAmount = getattr(market_pair.taker.market, "network_transaction_fee")
                transaction_fee = transaction_fee.amount
                # Transaction fee in maker quote asset
                transaction_fee *= gas_rate

                # The remaining quantity of the maker order to potentially hedge against
                quantity_remaining = active_order.quantity
                if not active_order.filled_quantity.is_nan():
                    quantity_remaining = active_order.quantity - active_order.filled_quantity

                if active_order.is_buy:
                    # Maker buy
                    # Taker sell
                    hedged_order_quantity = min(
                        quantity_remaining * base_rate,
                        market_pair.taker.market.get_available_balance(market_pair.taker.base_asset) *
                        self.order_size_taker_balance_factor
                    )
                    # Convert from taker to maker order amount
                    hedged_order_quantity = hedged_order_quantity / base_rate
                    # Calculate P/L including potential gas fees (if gateway)
                    pl = hedged_order_quantity * current_hedging_price - \
                        quantity_remaining * active_order.price - \
                        transaction_fee
                else:
                    # Maker sell
                    # Taker buy
                    taker_price = await market_pair.taker.market.get_order_price(
                        market_pair.taker.trading_pair,
                        True,
                        quantity_remaining * base_rate
                    )
                    hedged_order_quantity = min(
                        quantity_remaining * base_rate,
                        market_pair.taker.market.get_available_balance(market_pair.taker.quote_asset) /
                        taker_price * self.order_size_taker_balance_factor
                    )
                    # Convert from taker to maker order amount
                    hedged_order_quantity = hedged_order_quantity / base_rate
                    # Calculate P/L including potential gas fees (if gateway)
                    pl = quantity_remaining * active_order.price - \
                        hedged_order_quantity * current_hedging_price - \
                        transaction_fee

        # Profitability based on a price multiplier (cancel_order_threshold)
        # Profitability based on absolute P/L including fees
        if ((is_buy and current_hedging_price < order_price * (1 + cancel_order_threshold)) or
                (not is_buy and order_price < current_hedging_price * (1 + cancel_order_threshold)) or
                pl < 0):

            if LogOption.REMOVING_ORDER in self.logging_options:
                self.log_with_clock(
                    logging.INFO,
                    f"({market_pair.maker.trading_pair}) Limit {limit_order_type_str} order at "
                    f"{order_price:.8g} {market_pair.maker.quote_asset} is no longer profitable. "
                    f"Removing the order."
                )
            self.cancel_maker_order(market_pair, active_order.client_order_id)
            return False
        return True

    async def check_if_sufficient_balance(self, market_pair: MakerTakerMarketPair, active_order: LimitOrder) -> bool:
        """
        Check whether there's enough asset balance for a currently active limit order. If there's not enough asset
        balance for the order (e.g. because the required asset has been moved), cancel the active order.

        This function is only used when active order canceled is enabled.

        :param market_pair: cross exchange market pair
        :param active_order: current maker limit order
        :return: True if there's sufficient balance for the limit order, False if there isn't and the order is being
                 canceled.
        """
        is_buy = active_order.is_buy
        order_price = active_order.price
        # Maker order size
        size = self.get_adjusted_limit_order_size(market_pair)
        maker_market = market_pair.maker.market
        taker_trading_pair = market_pair.taker.trading_pair

        # Convert maker order size (in maker base asset) to taker order size (in taker base asset)
        _, _, quote_rate, _, _, base_rate, _, _, gas_rate = self.get_conversion_rates(market_pair)
        size /= base_rate

        taker_market = market_pair.taker.market
        quote_pair, quote_rate_source, quote_rate, base_pair, base_rate_source, base_rate, gas_pair, gas_rate_source, gas_rate = \
            self.get_conversion_rates(market_pair)

        if is_buy:
            # Maker buy
            # Taker sell
            taker_slippage_adjustment_factor = Decimal("1") - self.slippage_buffer

            quote_asset_amount = maker_market.get_balance(market_pair.maker.quote_asset)
            base_asset_amount = taker_market.get_balance(market_pair.taker.base_asset) * base_rate

            order_size_limit = min(base_asset_amount, quote_asset_amount / order_price)
        else:
            # Maker sell
            # Taker buy
            taker_slippage_adjustment_factor = Decimal("1") + self.slippage_buffer

            base_asset_amount = maker_market.get_balance(market_pair.maker.base_asset)
            quote_asset_amount = taker_market.get_balance(market_pair.taker.quote_asset)

            if self.is_gateway_market(market_pair.taker):
                taker_price = await taker_market.get_order_price(taker_trading_pair,
                                                                 True,
                                                                 size)
                if taker_price is None:
                    self.logger().warning("Gateway: failed to obtain order price."
                                          "Failed to determine sufficient balance.")
                    return False
            else:
                taker_price = taker_market.get_price_for_quote_volume(
                    taker_trading_pair, True, quote_asset_amount
                ).result_price

            adjusted_taker_price = (taker_price / base_rate) * taker_slippage_adjustment_factor
            order_size_limit = min(base_asset_amount, quote_asset_amount / adjusted_taker_price)

        quantized_size_limit = maker_market.quantize_order_amount(active_order.trading_pair, order_size_limit)

        if active_order.quantity > quantized_size_limit:
            if LogOption.ADJUST_ORDER in self.logging_options:
                self.log_with_clock(
                    logging.INFO,
                    f"({market_pair.maker.trading_pair}) Order size limit ({order_size_limit:.8g}) "
                    f"is now less than the current active order amount ({active_order.quantity:.8g}). "
                    f"Going to adjust the order."
                )
            self.cancel_maker_order(market_pair, active_order.client_order_id)
            return False
        return True

    def markettaker_to_maker_base_conversion_rate(self, market_pair: MarketTradingPairTuple) -> Decimal:
        """
        Return price conversion rate for from taker quote asset to maker quote asset
        """
        _, _, quote_rate, _, _, base_rate, _, _, _ = self.get_conversion_rates(market_pair)
        return quote_rate / base_rate
        # else:
        #     market_pairs = list(self._market_pairs.values())[0]
        #     quote_pair = f"{market_pairs.taker.quote_asset}-{market_pairs.maker.quote_asset}"
        #     base_pair = f"{market_pairs.taker.base_asset}-{market_pairs.maker.base_asset}"
        #     quote_rate = RateOracle.get_instance().rate(quote_pair)
        #     base_rate = RateOracle.get_instance().rate(base_pair)
        #     return quote_rate / base_rate

    async def check_and_create_new_orders(self,
                                          market_pair: MarketTradingPairTuple,
                                          has_active_bid: bool,
                                          has_active_ask: bool):
        """
        Check and account for all applicable conditions for creating new limit orders (e.g. profitability, what's the
        right price given depth tolerance and transient orders on the market, account balances, etc.), and create new
        limit orders for market making.

        :param market_pair: cross exchange market pair
        :param has_active_bid: True if there's already an active bid on the maker side, False otherwise
        :param has_active_ask: True if there's already an active ask on the maker side, False otherwise
        """

        # if there is no active bid, place bid again
        if not has_active_bid:
            bid_size = await self.get_market_making_size(market_pair, True)

            if bid_size > s_decimal_zero:
                bid_price = await self.get_market_making_price(market_pair, True, bid_size)
                if not Decimal.is_nan(bid_price):
                    effective_hedging_price = await self.calculate_effective_hedging_price(
                        market_pair,
                        True,
                        bid_size
                    )
                    effective_hedging_price_adjusted = effective_hedging_price / \
                        self.markettaker_to_maker_base_conversion_rate(market_pair)
                    if LogOption.CREATE_ORDER in self.logging_options:
                        self.log_with_clock(
                            logging.INFO,
                            f"({market_pair.maker.trading_pair}) Creating limit bid order for "
                            f"{bid_size} {market_pair.maker.base_asset} at "
                            f"{bid_price} {market_pair.maker.quote_asset}. "
                            f"Current hedging price: {effective_hedging_price:.8f} {market_pair.maker.quote_asset} "
                            f"(Rate adjusted: {effective_hedging_price_adjusted:.8f} {market_pair.taker.quote_asset})."
                        )
                    self.place_order(market_pair, True, True, bid_size, bid_price)
                else:
                    if LogOption.NULL_ORDER_SIZE in self.logging_options:
                        self.log_with_clock(
                            logging.WARNING,
                            f"({market_pair.maker.trading_pair})"
                            f"Order book on taker is too thin to place order for size: {bid_size}"
                            f"Reduce order_size_portfolio_ratio_limit"
                        )
            else:
                if LogOption.NULL_ORDER_SIZE in self.logging_options:
                    self.log_with_clock(
                        logging.WARNING,
                        f"({market_pair.maker.trading_pair}) Attempting to place a limit bid but the "
                        f"bid size is 0. Skipping. Check available balance."
                    )
        # if there is no active ask, place ask again
        if not has_active_ask:
            ask_size = await self.get_market_making_size(market_pair, False)

            if ask_size > s_decimal_zero:
                ask_price = await self.get_market_making_price(market_pair, False, ask_size)
                if not Decimal.is_nan(ask_price):
                    effective_hedging_price = await self.calculate_effective_hedging_price(
                        market_pair,
                        False,
                        ask_size
                    )
                    effective_hedging_price_adjusted = effective_hedging_price / \
                        self.markettaker_to_maker_base_conversion_rate(market_pair)
                    if LogOption.CREATE_ORDER in self.logging_options:
                        self.log_with_clock(
                            logging.INFO,
                            f"({market_pair.maker.trading_pair}) Creating limit ask order for "
                            f"{ask_size} {market_pair.maker.base_asset} at "
                            f"{ask_price} {market_pair.maker.quote_asset}. "
                            f"Current hedging price: {effective_hedging_price:.8f} {market_pair.maker.quote_asset} "
                            f"(Rate adjusted: {effective_hedging_price_adjusted:.8f} {market_pair.taker.quote_asset})."
                        )
                    self.place_order(market_pair, False, True, ask_size, ask_price)
                else:
                    if LogOption.NULL_ORDER_SIZE in self.logging_options:
                        self.log_with_clock(
                            logging.WARNING,
                            f"({market_pair.maker.trading_pair})"
                            f"Order book on taker is too thin to place order for size: {ask_size}"
                            f"Reduce order_size_portfolio_ratio_limit"
                        )
            else:
                if LogOption.NULL_ORDER_SIZE in self.logging_options:
                    self.log_with_clock(
                        logging.WARNING,
                        f"({market_pair.maker.trading_pair}) Attempting to place a limit ask but the "
                        f"ask size is 0. Skipping. Check available balance."
                    )

    def place_order(
        self,
        market_pair: MakerTakerMarketPair,
        is_buy: bool,
        is_maker: bool,  # True for maker order, False for taker order
        amount: Decimal,
        price: Decimal,
        maker_order_id: str = None,
        fill_records: List[OrderFilledEvent] = None,
    ):
        expiration_seconds = s_float_nan
        market_info = market_pair.maker if is_maker else market_pair.taker
        # Market orders are not being submitted as taker orders, limit orders are preferred at all times
        order_type = market_info.market.get_maker_order_type() if is_maker else \
            OrderType.LIMIT
        if order_type is OrderType.MARKET:
            price = s_decimal_nan
        expiration_seconds = self._config_map.order_refresh_mode.get_expiration_seconds()
        order_id = None
        if is_buy:
            try:
                order_id = self.buy_with_specific_market(market_info, amount,
                                                         order_type=order_type, price=price,
                                                         expiration_seconds=expiration_seconds)
            except ValueError as e:
                self.logger().warning(f"Placing an order on market {str(market_info.market.name)} "
                                      f"failed with the following error: {str(e)}")
        else:
            try:
                order_id = self.sell_with_specific_market(market_info, amount,
                                                          order_type=order_type, price=price,
                                                          expiration_seconds=expiration_seconds)
            except ValueError as e:
                self.logger().warning(f"Placing an order on market {str(market_info.market.name)} "
                                      f"failed with the following error: {str(e)}")
        if order_id is None:
            return
        self._sb_order_tracker.add_create_order_pending(order_id)
        self._market_pair_tracker.start_tracking_order_id(order_id, market_info.market, market_pair)
        if is_maker:
            self._maker_to_taker_order_ids[order_id] = []
        else:
            self._taker_to_maker_order_ids[order_id] = maker_order_id
            self._maker_to_taker_order_ids[maker_order_id] += [order_id]
            self.set_ongoing_hedging(fill_records, order_id)
        return order_id

    def cancel_maker_order(self, market_pair: MakerTakerMarketPair, order_id: str):
        market_trading_pair_tuple = self._market_pair_tracker.get_market_pair_from_order_id(order_id)
        super().cancel_order(market_trading_pair_tuple.maker, order_id)
    # ----------------------------------------------------------------------------------------------------------
    # </editor-fold>

    # <editor-fold desc="+ Order tracking entry points">
    # Override the stop tracking entry points to include the market pair tracker as well.
    # ----------------------------------------------------------------------------------------------------------
    def stop_tracking_limit_order(self, market_trading_pair_tuple, order_id: str):
        self._market_pair_tracker.stop_tracking_order_id(order_id)
        self.stop_tracking_limit_order(self, market_trading_pair_tuple, order_id)

    def stop_tracking_market_order(self, market_trading_pair_tuple, order_id: str):
        self._market_pair_tracker.stop_tracking_order_id(order_id)
        self.stop_tracking_market_order(self, market_trading_pair_tuple, order_id)
    # ----------------------------------------------------------------------------------------------------------
    # </editor-fold>

    # Removes orders from pending_create
    def did_create_buy_order(self, order_created_event):
        order_id = order_created_event.order_id
        self._sb_order_tracker.remove_create_order_pending(order_id)

    def did_create_sell_order(self, order_created_event):
        order_id = order_created_event.order_id
        self._sb_order_tracker.remove_create_order_pending(order_id)

    def notify_hb_app(self, msg: str):
        if self._hb_app_notification:
            super().notify_hb_app(msg)

    # ----------------------------------------------------------------------------------------------------------
    # Helpers
    def check_multiple_buy_orders(self, fill_records: List[OrderFilledEvent]):
        maker_order_ids = [r.order_id for _, r in fill_records]
        if len(set(maker_order_ids)) != 1:
            self.logger().warning("Multiple buy maker orders")

    def check_multiple_sell_orders(self, fill_records: List[OrderFilledEvent]):
        maker_order_ids = [r.order_id for _, r in fill_records]
        if len(set(maker_order_ids)) != 1:
            self.logger().warning("Multiple sell maker orders")

    def get_unhedged_buy_records(self, market_pair: MakerTakerMarketPair) -> List[OrderFilledEvent]:
        buy_fill_records = self._order_fill_buy_events.get(market_pair, [])
        return self.get_unhedged_events(buy_fill_records)

    def get_unhedged_sell_records(self, market_pair: MakerTakerMarketPair) -> List[OrderFilledEvent]:
        sell_fill_records = self._order_fill_sell_events.get(market_pair, [])
        return self.get_unhedged_events(sell_fill_records)

    def get_unhedged_events(self, fill_records: List[OrderFilledEvent]) -> List[OrderFilledEvent]:
        return [
            fill_event for fill_event in fill_records if (
                not self.is_fill_event_in_ongoing_hedging(fill_event)
            )
        ]

    def is_fill_event_in_ongoing_hedging(self, fill_event: OrderFilledEvent) -> bool:
        trade_id = fill_event[1].exchange_trade_id
        for maker_exchange_trade_ids in self._ongoing_hedging:
            for id in maker_exchange_trade_ids:
                if trade_id is id:
                    return True
        return False

    def set_ongoing_hedging(self, fill_records: List[OrderFilledEvent], order_id: str):
        maker_exchange_trade_ids = tuple(r.exchange_trade_id for _, r in fill_records)
        self._ongoing_hedging[maker_exchange_trade_ids] = order_id

    def del_order_from_ongoing_hedging(self, taker_order_id: str):
        maker_exchange_trade_ids = self._ongoing_hedging.inverse[taker_order_id]
        del self._ongoing_hedging[maker_exchange_trade_ids]



from sqlalchemy import BigInteger, Column, Index, Text

from . import HummingbotBase
from .decimal_type_decorator import SqliteDecimal


class Position(HummingbotBase):
    """
    Database model for storing positions held by executors.
    """
    __tablename__ = "Position"
    __table_args__ = (Index("p_controller_id_timestamp_index",
                            "controller_id", "timestamp"),
                      Index("p_connector_name_trading_pair_timestamp_index",
                            "connector_name", "trading_pair", "timestamp"))

    id = Column(Text, primary_key=True, nullable=False)
    controller_id = Column(Text, nullable=False)
    connector_name = Column(Text, nullable=False)
    side = Column(Text, nullable=False)
    trading_pair = Column(Text, nullable=False)
    timestamp = Column(BigInteger, nullable=False)
    volume_traded_quote = Column(SqliteDecimal(6), nullable=False)
    amount = Column(SqliteDecimal(6), nullable=False)
    breakeven_price = Column(SqliteDecimal(6), nullable=False)
    unrealized_pnl_quote = Column(SqliteDecimal(6), nullable=False)
    cum_fees_quote = Column(SqliteDecimal(6), nullable=False)

    def __repr__(self) -> str:
        return (f"Position(id='{self.id}', controller_id='{self.controller_id}', "
                f"connector_name='{self.connector_name}', trading_pair='{self.trading_pair}', "
                f"trading_pair='{self.trading_pair}', timestamp={self.timestamp}, "
                f"volume_traded_quote={self.volume_traded_quote}, amount={self.amount}, "
                f"breakeven_price={self.breakeven_price}, unrealized_pnl_quote={self.unrealized_pnl_quote}, "
                f"cum_fees_quote={self.cum_fees_quote})")



from typing import Callable, Optional

import hummingbot.connector.exchange.bitrue.bitrue_constants as CONSTANTS
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.connector.utils import GZipCompressionWSPostProcessor, TimeSynchronizerRESTPreProcessor
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


def public_rest_url(path_url: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided public REST endpoint
    :param path_url: a public REST endpoint
    :param domain: the Binance domain to connect to ("com" or "us"). The default value is "com"
    :return: the full URL to the endpoint
    """
    return CONSTANTS.REST_URL.format(domain) + CONSTANTS.API_VERSION + path_url


def private_rest_url(path_url: str, domain: str = CONSTANTS.DEFAULT_DOMAIN) -> str:
    """
    Creates a full URL for provided private REST endpoint
    :param path_url: a private REST endpoint
    :param domain: the Binance domain to connect to ("com" or "us"). The default value is "com"
    :return: the full URL to the endpoint
    """
    return CONSTANTS.REST_URL.format(domain) + CONSTANTS.API_VERSION + path_url


def build_api_factory(
    throttler: Optional[AsyncThrottler] = None,
    time_synchronizer: Optional[TimeSynchronizer] = None,
    time_provider: Optional[Callable] = None,
    auth: Optional[AuthBase] = None,
) -> WebAssistantsFactory:
    throttler = throttler or create_throttler()
    time_synchronizer = time_synchronizer or TimeSynchronizer()
    time_provider = time_provider or (
        lambda: get_current_server_time(
            throttler=throttler,
        )
    )
    api_factory = WebAssistantsFactory(
        throttler=throttler,
        auth=auth,
        rest_pre_processors=[
            TimeSynchronizerRESTPreProcessor(synchronizer=time_synchronizer, time_provider=time_provider),
        ],
        ws_post_processors=[GZipCompressionWSPostProcessor()],
    )
    return api_factory


def build_api_factory_without_time_synchronizer_pre_processor(throttler: AsyncThrottler) -> WebAssistantsFactory:
    api_factory = WebAssistantsFactory(throttler=throttler)
    return api_factory


def create_throttler() -> AsyncThrottler:
    return AsyncThrottler(CONSTANTS.RATE_LIMITS)


async def get_current_server_time(
    throttler: Optional[AsyncThrottler] = None,
    domain: str = "",
) -> float:
    throttler = throttler or create_throttler()
    api_factory = build_api_factory_without_time_synchronizer_pre_processor(throttler=throttler)
    rest_assistant = await api_factory.get_rest_assistant()
    response = await rest_assistant.execute_request(
        url=public_rest_url(path_url=CONSTANTS.SERVER_TIME_PATH_URL),
        method=RESTMethod.GET,
        throttler_limit_id=CONSTANTS.SERVER_TIME_PATH_URL,
    )
    server_time = response["serverTime"]
    return server_time



from decimal import Decimal
from typing import Any, Dict

from pydantic import ConfigDict, Field, SecretStr

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap
from hummingbot.connector.utils import split_hb_trading_pair
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

# https://www.okx.com/es-la/fees/
DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0.0002"),
    taker_percent_fee_decimal=Decimal("0.0005"),
)

CENTRALIZED = True

EXAMPLE_PAIR = "BTC-USDT"


def is_exchange_information_valid(exchange_info: Dict[str, Any]) -> bool:
    """
    Verifies if a trading pair is enabled to operate with based on its exchange information

    :param exchange_info: the exchange information for a trading pair

    :return: True if the trading pair is enabled, False otherwise
    """
    return (exchange_info.get("instType") == "SWAP"
            and exchange_info.get("ctType") == "linear"
            and exchange_info.get("state") == "live")


def is_linear_perpetual(trading_pair: str) -> bool:
    """
    Returns True if trading_pair is in USDT(Linear) Perpetual
    """
    _, quote_asset = split_hb_trading_pair(trading_pair)
    return quote_asset in ["USDT", "USDC"]


def get_next_funding_timestamp(current_timestamp: float) -> float:
    # On Okx Perpetuals, funding occurs every 8 hours at 00:00UTC, 08:00UTC and 16:00UTC.
    # Reference: https://help.okx.com/hc/en-us/articles/360039261134-Funding-fee-calculation
    int_ts = int(current_timestamp)
    eight_hours = 8 * 60 * 60
    mod = int_ts % eight_hours
    return float(int_ts - mod + eight_hours)


class OkxPerpetualConfigMap(BaseConnectorConfigMap):
    connector: str = "okx_perpetual"
    okx_perpetual_api_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Okx Perpetual API key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    okx_perpetual_secret_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Okx Perpetual secret key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    okx_perpetual_passphrase: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Okx Perpetual passphrase",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    model_config = ConfigDict(title="okx_perpetual")


KEYS = OkxPerpetualConfigMap.model_construct()



import asyncio
from decimal import Decimal
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field

from hummingbot.connector.gateway.common_types import ConnectorType, get_connector_type
from hummingbot.connector.gateway.gateway_swap import GatewaySwap
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.utils import async_ttl_cache
from hummingbot.core.utils.async_utils import safe_ensure_future


class TokenInfo(BaseModel):
    address: str
    symbol: str
    decimals: int


class AMMPoolInfo(BaseModel):
    address: str
    base_token_address: str = Field(alias="baseTokenAddress")
    quote_token_address: str = Field(alias="quoteTokenAddress")
    price: float
    fee_pct: float = Field(alias="feePct")
    base_token_amount: float = Field(alias="baseTokenAmount")
    quote_token_amount: float = Field(alias="quoteTokenAmount")


class CLMMPoolInfo(BaseModel):
    address: str
    base_token_address: str = Field(alias="baseTokenAddress")
    quote_token_address: str = Field(alias="quoteTokenAddress")
    bin_step: int = Field(alias="binStep")
    fee_pct: float = Field(alias="feePct")
    price: float
    base_token_amount: float = Field(alias="baseTokenAmount")
    quote_token_amount: float = Field(alias="quoteTokenAmount")
    active_bin_id: int = Field(alias="activeBinId")


class AMMPositionInfo(BaseModel):
    pool_address: str = Field(alias="poolAddress")
    wallet_address: str = Field(alias="walletAddress")
    base_token_address: str = Field(alias="baseTokenAddress")
    quote_token_address: str = Field(alias="quoteTokenAddress")
    lp_token_amount: float = Field(alias="lpTokenAmount")
    base_token_amount: float = Field(alias="baseTokenAmount")
    quote_token_amount: float = Field(alias="quoteTokenAmount")
    price: float
    base_token: Optional[str] = None
    quote_token: Optional[str] = None


class CLMMPositionInfo(BaseModel):
    address: str
    pool_address: str = Field(alias="poolAddress")
    base_token_address: str = Field(alias="baseTokenAddress")
    quote_token_address: str = Field(alias="quoteTokenAddress")
    base_token_amount: float = Field(alias="baseTokenAmount")
    quote_token_amount: float = Field(alias="quoteTokenAmount")
    base_fee_amount: float = Field(alias="baseFeeAmount")
    quote_fee_amount: float = Field(alias="quoteFeeAmount")
    lower_bin_id: int = Field(alias="lowerBinId")
    upper_bin_id: int = Field(alias="upperBinId")
    lower_price: float = Field(alias="lowerPrice")
    upper_price: float = Field(alias="upperPrice")
    price: float
    base_token: Optional[str] = None
    quote_token: Optional[str] = None


class GatewayLp(GatewaySwap):
    """
    Handles AMM and CLMM liquidity provision functionality including fetching pool info and adding/removing liquidity.
    Maintains order tracking and wallet interactions in the base class.
    """

    async def get_pool_address(self, trading_pair: str) -> Optional[str]:
        """Get pool address for a trading pair"""
        try:
            self.logger().info(f"Fetching pool address for {trading_pair} on {self.connector_name}")

            # Parse connector to get type (amm or clmm)
            connector_type = get_connector_type(self.connector_name)
            pool_type = "clmm" if connector_type == ConnectorType.CLMM else "amm"

            # Get pool info from gateway using the get_pool method
            pool_info = await self._get_gateway_instance().get_pool(
                trading_pair=trading_pair,
                connector=self.connector_name.split("/")[0],  # Just the name part
                network=self.network,
                type=pool_type
            )

            pool_address = pool_info.get("address")
            if pool_address:
                self.logger().info(f"Pool address: {pool_address}")
            else:
                self.logger().warning(f"No pool address found for {trading_pair}")

            return pool_address

        except Exception as e:
            self.logger().error(f"Error getting pool address for {trading_pair}: {e}")
            return None

    @async_ttl_cache(ttl=5, maxsize=10)
    async def get_pool_info(
        self,
        trading_pair: str,
    ) -> Optional[Union[AMMPoolInfo, CLMMPoolInfo]]:
        """
        Retrieves pool information for a given trading pair.
        Uses the appropriate model (AMMPoolInfo or CLMMPoolInfo) based on connector type.
        """
        try:
            # First get the pool address for the trading pair
            pool_address = await self.get_pool_address(trading_pair)

            if not pool_address:
                self.logger().warning(f"Could not find pool address for {trading_pair}")
                return None

            resp: Dict[str, Any] = await self._get_gateway_instance().pool_info(
                connector=self.connector_name,
                network=self.network,
                pool_address=pool_address,
            )

            # Determine which model to use based on connector type
            connector_type = get_connector_type(self.connector_name)
            if connector_type == ConnectorType.CLMM:
                return CLMMPoolInfo(**resp) if resp else None
            elif connector_type == ConnectorType.AMM:
                return AMMPoolInfo(**resp) if resp else None
            else:
                self.logger().warning(f"Unknown connector type: {connector_type} for {self.connector_name}")
                return None

        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.logger().network(
                f"Error fetching pool info for {trading_pair} on {self.connector_name}.",
                exc_info=True,
                app_warning_msg=str(e)
            )
            return None

    def add_liquidity(self, trading_pair: str, price: float, **request_args) -> str:
        """
        Adds liquidity to a pool - either concentrated (CLMM) or regular (AMM) based on the connector type.
        :param trading_pair: The market trading pair
        :param price: The center price for the position.
        :param request_args: Additional arguments for liquidity addition
        :return: A newly created order id (internal).
        """
        trade_type: TradeType = TradeType.RANGE
        order_id: str = self.create_market_order_id(trade_type, trading_pair)

        # Check connector type and call appropriate function
        connector_type = get_connector_type(self.connector_name)
        if connector_type == ConnectorType.CLMM:
            safe_ensure_future(self._clmm_add_liquidity(trade_type, order_id, trading_pair, price, **request_args))
        elif connector_type == ConnectorType.AMM:
            safe_ensure_future(self._amm_add_liquidity(trade_type, order_id, trading_pair, price, **request_args))
        else:
            raise ValueError(f"Connector type {connector_type} does not support liquidity provision")

        return order_id

    async def _clmm_add_liquidity(
        self,
        trade_type: TradeType,
        order_id: str,
        trading_pair: str,
        price: float,
        lower_price: Optional[float] = None,
        upper_price: Optional[float] = None,
        upper_width_pct: Optional[float] = None,
        lower_width_pct: Optional[float] = None,
        base_token_amount: Optional[float] = None,
        quote_token_amount: Optional[float] = None,
        slippage_pct: Optional[float] = None,
    ):
        """
        Opens a concentrated liquidity position with explicit price range or calculated from percentages.

        :param trade_type: The trade type (should be RANGE)
        :param order_id: Internal order id (also called client_order_id)
        :param trading_pair: The trading pair for the position
        :param price: The center price for the position (used if lower/upper_price not provided)
        :param lower_price: Explicit lower price bound (takes priority over percentages)
        :param upper_price: Explicit upper price bound (takes priority over percentages)
        :param upper_width_pct: The upper range width percentage from center price (e.g. 10.0 for +10%)
        :param lower_width_pct: The lower range width percentage from center price (e.g. 5.0 for -5%)
        :param base_token_amount: Amount of base token to add (optional)
        :param quote_token_amount: Amount of quote token to add (optional)
        :param slippage_pct: Maximum allowed slippage percentage
        :return: Response from the gateway API
        """
        # Check connector type is CLMM
        if get_connector_type(self.connector_name) != ConnectorType.CLMM:
            raise ValueError(f"Connector {self.connector_name} is not of type CLMM.")

        # Split trading_pair to get base and quote tokens
        tokens = trading_pair.split("-")
        if len(tokens) != 2:
            raise ValueError(f"Invalid trading pair format: {trading_pair}")

        base_token, quote_token = tokens

        # Calculate the total amount in base token units
        base_amount = base_token_amount or 0.0
        quote_amount_in_base = (quote_token_amount or 0.0) / price if price > 0 else 0.0
        total_amount_in_base = base_amount + quote_amount_in_base

        # Start tracking order with calculated amount
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=trade_type,
                                  price=Decimal(str(price)),
                                  amount=Decimal(str(total_amount_in_base)))

        # Determine position price range
        # Priority: explicit prices > width percentages
        if lower_price is not None and upper_price is not None:
            # Use explicit price bounds (highest priority)
            pass  # lower_price and upper_price already set
        elif upper_width_pct is not None and lower_width_pct is not None:
            # Calculate from width percentages
            lower_width_decimal = lower_width_pct / 100.0
            upper_width_decimal = upper_width_pct / 100.0
            lower_price = price * (1 - lower_width_decimal)
            upper_price = price * (1 + upper_width_decimal)
        else:
            raise ValueError("Must provide either (lower_price and upper_price) or (upper_width_pct and lower_width_pct)")

        # Get pool address for the trading pair
        pool_address = await self.get_pool_address(trading_pair)
        if not pool_address:
            raise ValueError(f"Could not find pool for {trading_pair}")

        # Open position
        try:
            transaction_result = await self._get_gateway_instance().clmm_open_position(
                connector=self.connector_name,
                network=self.network,
                wallet_address=self.address,
                pool_address=pool_address,
                lower_price=lower_price,
                upper_price=upper_price,
                base_token_amount=base_token_amount,
                quote_token_amount=quote_token_amount,
                slippage_pct=slippage_pct
            )
            transaction_hash: Optional[str] = transaction_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, transaction_result)
                return transaction_hash
            else:
                raise ValueError("No transaction hash returned from gateway")
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, "opening CLMM position", e)

    async def _amm_add_liquidity(
        self,
        trade_type: TradeType,
        order_id: str,
        trading_pair: str,
        price: float,
        base_token_amount: float,
        quote_token_amount: float,
        slippage_pct: Optional[float] = None,
    ):
        """
        Opens a regular AMM liquidity position.

        :param trade_type: The trade type (should be RANGE)
        :param order_id: Internal order id (also called client_order_id)
        :param trading_pair: The trading pair for the position
        :param price: The price for the position
        :param base_token_amount: Amount of base token to add (required)
        :param quote_token_amount: Amount of quote token to add (required)
        :param slippage_pct: Maximum allowed slippage percentage
        """
        # Check connector type is AMM
        if get_connector_type(self.connector_name) != ConnectorType.AMM:
            raise ValueError(f"Connector {self.connector_name} is not of type AMM.")

        # Split trading_pair to get base and quote tokens
        tokens = trading_pair.split("-")
        if len(tokens) != 2:
            raise ValueError(f"Invalid trading pair format: {trading_pair}")

        base_token, quote_token = tokens

        # Calculate the total amount in base token units
        quote_amount_in_base = quote_token_amount / price if price > 0 else 0.0
        total_amount_in_base = base_token_amount + quote_amount_in_base

        # Start tracking order with calculated amount
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=trade_type,
                                  price=Decimal(str(price)),
                                  amount=Decimal(str(total_amount_in_base)))

        # Get pool address for the trading pair
        pool_address = await self.get_pool_address(trading_pair)
        if not pool_address:
            raise ValueError(f"Could not find pool for {trading_pair}")

        # Add liquidity to AMM pool
        try:
            transaction_result = await self._get_gateway_instance().amm_add_liquidity(
                connector=self.connector_name,
                network=self.network,
                wallet_address=self.address,
                pool_address=pool_address,
                base_token_amount=base_token_amount,
                quote_token_amount=quote_token_amount,
                slippage_pct=slippage_pct
            )
            transaction_hash: Optional[str] = transaction_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, transaction_result)
                return transaction_hash
            else:
                raise ValueError("No transaction hash returned from gateway")
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, "opening AMM position", e)

    def remove_liquidity(
        self,
        trading_pair: str,
        position_address: Optional[str] = None,
        percentage: float = 100.0,
        **request_args
    ) -> str:
        """
        Removes liquidity from a position - either concentrated (CLMM) or regular (AMM) based on the connector type.
        :param trading_pair: The market trading pair
        :param position_address: The address of the position (required for CLMM, optional for AMM)
        :param percentage: Percentage of liquidity to remove (defaults to 100%)
        :return: A newly created order id (internal).
        """
        connector_type = get_connector_type(self.connector_name)

        # Verify we have a position address for CLMM positions
        if connector_type == ConnectorType.CLMM and position_address is None:
            raise ValueError("position_address is required to close a CLMM position")

        trade_type: TradeType = TradeType.RANGE
        order_id: str = self.create_market_order_id(trade_type, trading_pair)

        # Call appropriate function based on connector type and percentage
        if connector_type == ConnectorType.CLMM:
            if percentage == 100.0:
                # Complete close for CLMM
                safe_ensure_future(self._clmm_close_position(trade_type, order_id, trading_pair, position_address, **request_args))
            else:
                # Partial removal for CLMM
                safe_ensure_future(self._clmm_remove_liquidity(trade_type, order_id, trading_pair, position_address, percentage, **request_args))
        elif connector_type == ConnectorType.AMM:
            # AMM always uses remove_liquidity
            safe_ensure_future(self._amm_remove_liquidity(trade_type, order_id, trading_pair, percentage, **request_args))
        else:
            raise ValueError(f"Connector type {connector_type} does not support liquidity provision")

        return order_id

    async def _clmm_close_position(
        self,
        trade_type: TradeType,
        order_id: str,
        trading_pair: str,
        position_address: str,
        fail_silently: bool = False,
    ):
        """
        Closes a concentrated liquidity position for the given position address.

        :param trade_type: The trade type (should be RANGE)
        :param order_id: Internal order id (also called client_order_id)
        :param trading_pair: The trading pair for the position
        :param position_address: The address of the position to close
        :param fail_silently: Whether to fail silently on error
        """
        # Check connector type is CLMM
        if get_connector_type(self.connector_name) != ConnectorType.CLMM:
            raise ValueError(f"Connector {self.connector_name} is not of type CLMM.")

        # Start tracking order
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=trade_type)
        try:
            transaction_result = await self._get_gateway_instance().clmm_close_position(
                connector=self.connector_name,
                network=self.network,
                wallet_address=self.address,
                position_address=position_address,
                fail_silently=fail_silently
            )
            transaction_hash: Optional[str] = transaction_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, transaction_result)
                return transaction_hash
            else:
                raise ValueError("No transaction hash returned from gateway")
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, "closing CLMM position", e)

    async def _clmm_remove_liquidity(
        self,
        trade_type: TradeType,
        order_id: str,
        trading_pair: str,
        position_address: str,
        percentage: float = 100.0,
        fail_silently: bool = False,
    ):
        """
        Removes liquidity from a CLMM position (partial removal).

        :param trade_type: The trade type (should be RANGE)
        :param order_id: Internal order id (also called client_order_id)
        :param trading_pair: The trading pair for the position
        :param position_address: The address of the position
        :param percentage: Percentage of liquidity to remove (0-100)
        :param fail_silently: Whether to fail silently on error
        """
        # Check connector type is CLMM
        if get_connector_type(self.connector_name) != ConnectorType.CLMM:
            raise ValueError(f"Connector {self.connector_name} is not of type CLMM.")

        # Start tracking order
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=trade_type)
        try:
            transaction_result = await self._get_gateway_instance().clmm_remove_liquidity(
                connector=self.connector_name,
                network=self.network,
                wallet_address=self.address,
                position_address=position_address,
                percentage=percentage,
                fail_silently=fail_silently
            )
            transaction_hash: Optional[str] = transaction_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, transaction_result)
                return transaction_hash
            else:
                raise ValueError("No transaction hash returned from gateway")
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, "removing CLMM liquidity", e)

    async def _amm_remove_liquidity(
        self,
        trade_type: TradeType,
        order_id: str,
        trading_pair: str,
        percentage: float = 100.0,
        fail_silently: bool = False,
    ):
        """
        Closes an AMM liquidity position by removing specified percentage of liquidity.

        :param trade_type: The trade type (should be RANGE)
        :param order_id: Internal order id (also called client_order_id)
        :param trading_pair: The trading pair for the position
        :param percentage: Percentage of liquidity to remove (0-100)
        :param fail_silently: Whether to fail silently on error
        """
        # Check connector type is AMM
        if get_connector_type(self.connector_name) != ConnectorType.AMM:
            raise ValueError(f"Connector {self.connector_name} is not of type AMM.")

        # Get pool address for the trading pair
        pool_address = await self.get_pool_address(trading_pair)
        if not pool_address:
            raise ValueError(f"Could not find pool for {trading_pair}")

        # Start tracking order
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=trade_type)

        try:
            transaction_result = await self._get_gateway_instance().amm_remove_liquidity(
                connector=self.connector_name,
                network=self.network,
                wallet_address=self.address,
                pool_address=pool_address,
                percentage=percentage,
                fail_silently=fail_silently
            )
            transaction_hash: Optional[str] = transaction_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, transaction_result)
                return transaction_hash
            else:
                raise ValueError("No transaction hash returned from gateway")
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, "closing AMM position", e)

    async def amm_add_liquidity(
        self,
        trading_pair: str,
        base_token_amount: float,
        quote_token_amount: float,
        slippage_pct: Optional[float] = None,
        fail_silently: bool = False
    ) -> Dict[str, Any]:
        """
        Adds liquidity to an AMM pool.
        :param trading_pair: The trading pair for the position
        :param base_token_amount: Amount of base token to add
        :param quote_token_amount: Amount of quote token to add
        :param slippage_pct: Maximum allowed slippage percentage
        :param fail_silently: Whether to fail silently on error
        :return: Response from the gateway API
        """
        # Check connector type is AMM
        if get_connector_type(self.connector_name) != ConnectorType.AMM:
            raise ValueError(f"Connector {self.connector_name} is not of type AMM.")

        # Get pool address for the trading pair
        pool_address = await self.get_pool_address(trading_pair)
        if not pool_address:
            raise ValueError(f"Could not find pool for {trading_pair}")

        order_id: str = self.create_market_order_id(TradeType.RANGE, trading_pair)
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=TradeType.RANGE)
        try:
            transaction_result = await self._get_gateway_instance().amm_add_liquidity(
                connector=self.connector_name,
                network=self.network,
                wallet_address=self.address,
                pool_address=pool_address,
                base_token_amount=base_token_amount,
                quote_token_amount=quote_token_amount,
                slippage_pct=slippage_pct,
                fail_silently=fail_silently
            )
            transaction_hash: Optional[str] = transaction_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, transaction_result)
            else:
                raise ValueError
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, "adding AMM liquidity", e)

    async def amm_remove_liquidity(
        self,
        trading_pair: str,
        percentage: float,
        fail_silently: bool = False
    ) -> Dict[str, Any]:
        """
        Removes liquidity from an AMM pool.
        :param trading_pair: The trading pair for the position
        :param percentage: Percentage of liquidity to remove (0-100)
        :param fail_silently: Whether to fail silently on error
        :return: Response from the gateway API
        """
        # Check connector type is AMM
        if get_connector_type(self.connector_name) != ConnectorType.AMM:
            raise ValueError(f"Connector {self.connector_name} is not of type AMM.")

        # Get pool address for the trading pair
        pool_address = await self.get_pool_address(trading_pair)
        if not pool_address:
            raise ValueError(f"Could not find pool for {trading_pair}")

        order_id: str = self.create_market_order_id(TradeType.RANGE, trading_pair)
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=TradeType.RANGE)
        try:
            transaction_result = await self._get_gateway_instance().amm_remove_liquidity(
                connector=self.connector_name,
                network=self.network,
                wallet_address=self.address,
                pool_address=pool_address,
                percentage=percentage,
                fail_silently=fail_silently
            )
            transaction_hash: Optional[str] = transaction_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, transaction_result)
            else:
                raise ValueError
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, "removing AMM liquidity", e)

    @async_ttl_cache(ttl=5, maxsize=10)
    async def get_position_info(
        self,
        trading_pair: str,
        position_address: Optional[str] = None
    ) -> Union[AMMPositionInfo, CLMMPositionInfo, None]:
        """
        Retrieves position information for a given liquidity position.

        :param trading_pair: The trading pair for the position
        :param position_address: The address of the position (required for CLMM, optional for AMM)
        :return: Position information from gateway, validated against appropriate schema
        """
        try:
            # Split trading_pair to get base and quote tokens
            tokens = trading_pair.split("-")
            if len(tokens) != 2:
                raise ValueError(f"Invalid trading pair format: {trading_pair}")

            base_token, quote_token = tokens

            connector_type = get_connector_type(self.connector_name)
            if connector_type == ConnectorType.CLMM:
                if position_address is None:
                    raise ValueError("position_address is required for CLMM positions")

                resp: Dict[str, Any] = await self._get_gateway_instance().clmm_position_info(
                    connector=self.connector_name,
                    network=self.network,
                    position_address=position_address,
                    wallet_address=self.address,
                )
                # Validate response against CLMM schema
                return CLMMPositionInfo(**resp) if resp else None

            elif connector_type == ConnectorType.AMM:
                resp: Dict[str, Any] = await self._get_gateway_instance().amm_position_info(
                    connector=self.connector_name,
                    network=self.network,
                    pool_address=position_address,
                    wallet_address=self.address,
                )
                # Validate response against AMM schema
                return AMMPositionInfo(**resp) if resp else None

            else:
                raise ValueError(f"Connector type {connector_type} does not support liquidity positions")

        except asyncio.CancelledError:
            raise
        except Exception as e:
            addr_info = f"position {position_address}" if position_address else trading_pair
            self.logger().network(
                f"Error fetching position info for {addr_info} on {self.connector_name}.",
                exc_info=True,
                app_warning_msg=str(e)
            )
            return None

    async def get_user_positions(self, pool_address: Optional[str] = None) -> List[Union[AMMPositionInfo, CLMMPositionInfo]]:
        """
        Fetch all user positions for this connector and wallet.

        :param pool_address: Optional pool address to filter positions (required for AMM)
        :return: List of position information objects
        """
        positions = []

        try:
            connector_type = get_connector_type(self.connector_name)

            if connector_type == ConnectorType.CLMM:
                # For CLMM, use positions-owned endpoint
                response = await self._get_gateway_instance().clmm_positions_owned(
                    connector=self.connector_name,
                    network=self.network,
                    wallet_address=self.address,
                    pool_address=pool_address  # Optional filter by pool
                )
            else:
                # For AMM, we need a pool address
                if not pool_address:
                    self.logger().warning("AMM position fetching requires a pool address")
                    return []

                # For AMM, get position info directly from the pool
                # We'll need to get pool info first to extract tokens
                pool_resp = await self._get_gateway_instance().pool_info(
                    connector=self.connector_name,
                    network=self.network,
                    pool_address=pool_address
                )

                if not pool_resp:
                    return []

                # Now get the position info
                resp = await self._get_gateway_instance().amm_position_info(
                    connector=self.connector_name,
                    network=self.network,
                    pool_address=pool_address,
                    wallet_address=self.address
                )

                if resp:
                    position = AMMPositionInfo(**resp)
                    # Get token symbols from loaded token data
                    base_token_info = self.get_token_by_address(position.base_token_address)
                    quote_token_info = self.get_token_by_address(position.quote_token_address)

                    # Use symbol if found, otherwise use address
                    position.base_token = base_token_info.get("symbol", position.base_token_address) if base_token_info else position.base_token_address
                    position.quote_token = quote_token_info.get("symbol", position.quote_token_address) if quote_token_info else position.quote_token_address
                    return [position]
                else:
                    return []

            # Parse position data based on connector type (for CLMM)
            # Handle case where response might be a list directly or a dict with 'positions' key
            positions_list = response if isinstance(response, list) else response.get("positions", [])
            for pos_data in positions_list:
                try:
                    if connector_type == ConnectorType.CLMM:
                        position = CLMMPositionInfo(**pos_data)

                        # Get token symbols from loaded token data
                        base_token_info = self.get_token_by_address(position.base_token_address)
                        quote_token_info = self.get_token_by_address(position.quote_token_address)

                        # Use symbol if found, otherwise use address
                        position.base_token = base_token_info.get("symbol", position.base_token_address) if base_token_info else position.base_token_address
                        position.quote_token = quote_token_info.get("symbol", position.quote_token_address) if quote_token_info else position.quote_token_address

                        positions.append(position)
                    else:
                        position = AMMPositionInfo(**pos_data)

                        # Get token symbols from loaded token data
                        base_token_info = self.get_token_by_address(position.base_token_address)
                        quote_token_info = self.get_token_by_address(position.quote_token_address)

                        # Use symbol if found, otherwise use address
                        position.base_token = base_token_info.get("symbol", position.base_token_address) if base_token_info else position.base_token_address
                        position.quote_token = quote_token_info.get("symbol", position.quote_token_address) if quote_token_info else position.quote_token_address

                        positions.append(position)

                except Exception as e:
                    self.logger().error(f"Error parsing position data: {e}", exc_info=True)
                    continue

        except Exception as e:
            self.logger().error(f"Error fetching positions: {e}", exc_info=True)

        return positions



import logging
from abc import ABC, abstractmethod

import numpy as np

from ..ring_buffer import RingBuffer

pmm_logger = None


class BaseTrailingIndicator(ABC):
    @classmethod
    def logger(cls):
        global pmm_logger
        if pmm_logger is None:
            pmm_logger = logging.getLogger(__name__)
        return pmm_logger

    def __init__(self, sampling_length: int = 30, processing_length: int = 15):
        self._sampling_buffer = RingBuffer(sampling_length)
        self._processing_buffer = RingBuffer(processing_length)
        self._samples_length = 0

    def add_sample(self, value: float):
        self._sampling_buffer.add_value(value)
        indicator_value = self._indicator_calculation()
        self._processing_buffer.add_value(indicator_value)

    @abstractmethod
    def _indicator_calculation(self) -> float:
        raise NotImplementedError

    def _processing_calculation(self) -> float:
        """
        Processing of the processing buffer to return final value.
        Default behavior is buffer average
        """
        return np.mean(self._processing_buffer.get_as_numpy_array())

    @property
    def current_value(self) -> float:
        return self._processing_calculation()

    @property
    def is_sampling_buffer_full(self) -> bool:
        return self._sampling_buffer.is_full

    @property
    def is_processing_buffer_full(self) -> bool:
        return self._processing_buffer.is_full

    @property
    def is_sampling_buffer_changed(self) -> bool:
        buffer_len = len(self._sampling_buffer.get_as_numpy_array())
        is_changed = self._samples_length != buffer_len
        self._samples_length = buffer_len
        return is_changed

    @property
    def sampling_length(self) -> int:
        return self._sampling_buffer.length

    @sampling_length.setter
    def sampling_length(self, value):
        self._sampling_buffer.length = value

    @property
    def processing_length(self) -> int:
        return self._processing_buffer.length

    @processing_length.setter
    def processing_length(self, value):
        self._processing_buffer.length = value



import asyncio
from decimal import Decimal
from typing import Any, Dict, List, Optional, Tuple

from bidict import bidict

import hummingbot.connector.derivative.bitget_perpetual.bitget_perpetual_constants as CONSTANTS
from hummingbot.connector.derivative.bitget_perpetual import (
    bitget_perpetual_utils,
    bitget_perpetual_web_utils as web_utils,
)
from hummingbot.connector.derivative.bitget_perpetual.bitget_perpetual_api_order_book_data_source import (
    BitgetPerpetualAPIOrderBookDataSource,
)
from hummingbot.connector.derivative.bitget_perpetual.bitget_perpetual_api_user_stream_data_source import (
    BitgetPerpetualUserStreamDataSource,
)
from hummingbot.connector.derivative.bitget_perpetual.bitget_perpetual_auth import BitgetPerpetualAuth
from hummingbot.connector.derivative.bitget_perpetual.bitget_perpetual_constants import MarginMode
from hummingbot.connector.derivative.position import Position
from hummingbot.connector.perpetual_derivative_py_base import PerpetualDerivativePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair, split_hb_trading_pair
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType, PositionAction, PositionMode, PositionSide, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase, TradeFeeSchema
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory

s_decimal_NaN = Decimal("nan")
s_decimal_0 = Decimal(0)


class BitgetPerpetualDerivative(PerpetualDerivativePyBase):

    web_utils = web_utils

    def __init__(
        self,
        balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
        rate_limits_share_pct: Decimal = Decimal("100"),
        bitget_perpetual_api_key: str = None,
        bitget_perpetual_secret_key: str = None,
        bitget_perpetual_passphrase: str = None,
        trading_pairs: Optional[List[str]] = None,
        trading_required: bool = True,
    ) -> None:

        self.bitget_perpetual_api_key = bitget_perpetual_api_key
        self.bitget_perpetual_secret_key = bitget_perpetual_secret_key
        self.bitget_perpetual_passphrase = bitget_perpetual_passphrase
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._last_trade_history_timestamp = None

        self._margin_mode = MarginMode.CROSS

        super().__init__(balance_asset_limit, rate_limits_share_pct)

    @property
    def name(self) -> str:
        return CONSTANTS.EXCHANGE_NAME

    @property
    def authenticator(self) -> BitgetPerpetualAuth:
        return BitgetPerpetualAuth(
            api_key=self.bitget_perpetual_api_key,
            secret_key=self.bitget_perpetual_secret_key,
            passphrase=self.bitget_perpetual_passphrase,
            time_provider=self._time_synchronizer
        )

    @property
    def rate_limits_rules(self) -> List[RateLimit]:
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self) -> str:
        return CONSTANTS.DEFAULT_DOMAIN

    @property
    def client_order_id_max_length(self) -> int:
        return CONSTANTS.ORDER_ID_MAX_LEN

    @property
    def client_order_id_prefix(self) -> str:
        return CONSTANTS.HBOT_ORDER_ID_PREFIX

    @property
    def trading_rules_request_path(self) -> str:
        return CONSTANTS.PUBLIC_CONTRACTS_ENDPOINT

    @property
    def trading_pairs_request_path(self) -> str:
        return CONSTANTS.PUBLIC_CONTRACTS_ENDPOINT

    @property
    def check_network_request_path(self) -> str:
        return CONSTANTS.PUBLIC_TIME_ENDPOINT

    @property
    def trading_pairs(self) -> Optional[List[str]]:
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return False

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    @property
    def funding_fee_poll_interval(self) -> int:
        return 120

    @staticmethod
    def _formatted_error(code: int, message: str) -> str:
        return f"Error: {code} - {message}"

    async def start_network(self):
        await super().start_network()
        await self.set_margin_mode(self._margin_mode)

        if self.is_trading_required:
            self.set_position_mode(PositionMode.HEDGE)

    def supported_order_types(self) -> List[OrderType]:
        return [OrderType.LIMIT, OrderType.MARKET]

    def supported_position_modes(self) -> List[PositionMode]:
        return [PositionMode.ONEWAY, PositionMode.HEDGE]

    def _is_request_exception_related_to_time_synchronizer(
        self,
        request_exception: Exception
    ) -> bool:
        error_description = str(request_exception)
        ts_error_target_str = "Request timestamp expired"

        return ts_error_target_str in error_description

    def _collateral_token_based_on_trading_pair(self, trading_pair: str) -> str:
        """
        Returns the collateral token based on the trading pair
        (For example this method need for order cancellation)

        :return: The collateral token
        """
        base, quote = split_hb_trading_pair(trading_pair=trading_pair)

        if quote == "USD":
            collateral_token = base
        else:
            collateral_token = quote

        return collateral_token

    async def get_exchange_position_mode(self, trading_pair: str) -> None:
        """
        Returns the current exchange position mode.
        """
        product_type = await self.product_type_associated_to_trading_pair(trading_pair)
        account_info_response: Dict[str, Any] = await self._api_get(
            path_url=CONSTANTS.ACCOUNT_INFO_ENDPOINT,
            params={
                "symbol": await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair),
                "productType": product_type,
                "marginCoin": self.get_buy_collateral_token(trading_pair),
            },
            is_auth_required=True,
        )
        if account_info_response["code"] != CONSTANTS.RET_CODE_OK:
            self.logger().error(self._formatted_error(
                account_info_response["code"],
                f"Error getting position mode for {trading_pair}: {account_info_response['msg']}"
            ))
            return

        position_modes = {
            "one_way_mode": PositionMode.ONEWAY,
            "hedge_mode": PositionMode.HEDGE,
        }

        position_mode = position_modes[account_info_response["data"]["posMode"]]

        self.logger().info(f"Position mode for {trading_pair}: {position_mode}")

    def get_buy_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules.get(trading_pair, None)
        if trading_rule is None:
            collateral_token = self._collateral_token_based_on_trading_pair(
                trading_pair=trading_pair
            )
        else:
            collateral_token = trading_rule.buy_order_collateral_token

        return collateral_token

    def get_sell_collateral_token(self, trading_pair: str) -> str:
        return self.get_buy_collateral_token(trading_pair=trading_pair)

    async def product_type_associated_to_trading_pair(self, trading_pair: str) -> str:
        """
        Returns the product type associated with the trading pair
        """
        _, quote = split_hb_trading_pair(trading_pair)

        if quote == "USDT":
            return CONSTANTS.USDT_PRODUCT_TYPE

        if quote == "USDC":
            return CONSTANTS.USDC_PRODUCT_TYPE

        return CONSTANTS.USD_PRODUCT_TYPE

    def _is_order_not_found_during_status_update_error(
        self,
        status_update_exception: Exception
    ) -> bool:
        # Error example:
        # { "code": "00000", "msg": "success", "requestTime": 1710327684832, "data": [] }

        if isinstance(status_update_exception, IOError):
            return any(
                value in str(status_update_exception)
                for value in CONSTANTS.RET_CODES_ORDER_NOT_EXISTS
            )

        if isinstance(status_update_exception, ValueError):
            return True

        return False

    def _is_order_not_found_during_cancelation_error(
        self,
        cancelation_exception: Exception
    ) -> bool:
        if isinstance(cancelation_exception, IOError):
            return any(
                value in str(cancelation_exception)
                for value in CONSTANTS.RET_CODES_ORDER_NOT_EXISTS
            )

        return False

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        symbol = await self.exchange_symbol_associated_to_pair(tracked_order.trading_pair)
        product_type = await self.product_type_associated_to_trading_pair(
            tracked_order.trading_pair
        )
        cancel_result = await self._api_post(
            path_url=CONSTANTS.CANCEL_ORDER_ENDPOINT,
            data={
                "symbol": symbol,
                "productType": product_type,
                "marginCoin": self.get_buy_collateral_token(tracked_order.trading_pair),
                "orderId": tracked_order.exchange_order_id
            },
            is_auth_required=True,
        )
        response_code = cancel_result["code"]

        if response_code != CONSTANTS.RET_CODE_OK:
            raise IOError(self._formatted_error(response_code, cancel_result["msg"]))

        return True

    async def _place_order(
        self,
        order_id: str,
        trading_pair: str,
        amount: Decimal,
        trade_type: TradeType,
        order_type: OrderType,
        price: Decimal,
        position_action: PositionAction = PositionAction.NIL,
        **kwargs,
    ) -> Tuple[str, float]:
        product_type = await self.product_type_associated_to_trading_pair(trading_pair)
        margin_modes = {
            MarginMode.CROSS: "crossed",
            MarginMode.ISOLATED: "isolated"
        }
        data = {
            "marginCoin": self.get_buy_collateral_token(trading_pair),
            "symbol": await self.exchange_symbol_associated_to_pair(trading_pair),
            "productType": product_type,
            "size": str(amount),
            "force": CONSTANTS.DEFAULT_TIME_IN_FORCE,
            "clientOid": order_id,
            "side": trade_type.name.lower(),
            "marginMode": margin_modes[self._margin_mode],
            "orderType": "limit" if order_type.is_limit_type() else "market",
        }
        if order_type.is_limit_type():
            data["price"] = str(price)

        if self.position_mode is PositionMode.HEDGE:
            if position_action is PositionAction.CLOSE:
                data["side"] = "sell" if trade_type is TradeType.BUY else "buy"
            data["tradeSide"] = position_action.name.lower()

        resp = await self._api_post(
            path_url=CONSTANTS.PLACE_ORDER_ENDPOINT,
            data=data,
            is_auth_required=True,
            headers={
                "X-CHANNEL-API-CODE": CONSTANTS.API_CODE,
            }
        )

        if resp["code"] != CONSTANTS.RET_CODE_OK:
            raise IOError(self._formatted_error(
                resp["code"],
                f"Error submitting order {order_id}: {resp['msg']}"
            ))

        return str(resp["data"]["orderId"]), self.current_timestamp

    def _get_fee(
        self,
        base_currency: str,
        quote_currency: str,
        order_type: OrderType,
        order_side: TradeType,
        position_action: PositionAction,
        amount: Decimal,
        price: Decimal = s_decimal_NaN,
        is_maker: Optional[bool] = None
    ) -> TradeFeeBase:
        is_maker = is_maker or (order_type is OrderType.LIMIT_MAKER)
        trading_pair = combine_to_hb_trading_pair(base=base_currency, quote=quote_currency)
        if trading_pair in self._trading_fees:
            fee_schema: TradeFeeSchema = self._trading_fees[trading_pair]
            fee_rate = (
                fee_schema.maker_percent_fee_decimal
                if is_maker
                else fee_schema.taker_percent_fee_decimal
            )
            fee = TradeFeeBase.new_spot_fee(
                fee_schema=fee_schema,
                trade_type=order_side,
                percent=fee_rate,
            )
        else:
            fee = build_trade_fee(
                self.name,
                is_maker,
                base_currency=base_currency,
                quote_currency=quote_currency,
                order_type=order_type,
                order_side=order_side,
                amount=amount,
                price=price,
            )
        return fee

    async def _update_trading_fees(self):
        symbol_data = []

        for product_type in CONSTANTS.ALL_PRODUCT_TYPES:
            exchange_info = await self._api_get(
                path_url=self.trading_rules_request_path,
                params={
                    "productType": product_type
                }
            )
            symbol_data.extend(exchange_info["data"])

        for symbol_details in symbol_data:
            if bitget_perpetual_utils.is_exchange_information_valid(exchange_info=symbol_details):
                trading_pair = await self.trading_pair_associated_to_exchange_symbol(
                    symbol=symbol_details["symbol"]
                )
                self._trading_fees[trading_pair] = TradeFeeSchema(
                    maker_percent_fee_decimal=Decimal(symbol_details["makerFeeRate"]),
                    taker_percent_fee_decimal=Decimal(symbol_details["takerFeeRate"])
                )

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            auth=self._auth,
        )

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return BitgetPerpetualAPIOrderBookDataSource(
            self.trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
        )

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return BitgetPerpetualUserStreamDataSource(
            auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
        )

    async def _update_balances(self):
        """
        Calls REST API to update total and available balances
        """
        balances = []
        product_types: set[str] = {
            await self.product_type_associated_to_trading_pair(trading_pair)
            for trading_pair in self._trading_pairs
        } or CONSTANTS.ALL_PRODUCT_TYPES

        for product_type in product_types:
            accounts_info_response: Dict[str, Any] = await self._api_get(
                path_url=CONSTANTS.ACCOUNTS_INFO_ENDPOINT,
                params={
                    "productType": product_type
                },
                is_auth_required=True,
            )

            if accounts_info_response["code"] != CONSTANTS.RET_CODE_OK:
                raise IOError(
                    self._formatted_error(
                        accounts_info_response["code"],
                        accounts_info_response["msg"]
                    )
                )

            balances.extend(accounts_info_response["data"])

        self._account_available_balances.clear()
        self._account_balances.clear()

        for balance_data in balances:
            quote_asset_name = balance_data["marginCoin"]
            queried_available = Decimal(balance_data["crossedMaxAvailable"])
            queried_total = Decimal(balance_data["accountEquity"])
            current_total = self._account_balances.get(quote_asset_name, Decimal(0))
            current_available = self._account_available_balances.get(quote_asset_name, Decimal(0))

            total = current_total + queried_total
            available = current_available + queried_available

            if total or available:
                self._account_available_balances[quote_asset_name] = available
                self._account_balances[quote_asset_name] = total

            if "assetList" in balance_data:
                for base_asset in balance_data["assetList"]:
                    base_asset_name = base_asset["coin"]
                    queried_available = Decimal(base_asset["available"])
                    queried_total = Decimal(base_asset["balance"])
                    current_total = self._account_balances.get(base_asset_name, Decimal(0))
                    current_available = self._account_available_balances.get(
                        base_asset_name,
                        Decimal(0)
                    )

                    total = current_total + queried_total
                    available = current_available + queried_available

                    if total or available:
                        self._account_available_balances[base_asset_name] = available
                        self._account_balances[base_asset_name] = total

    async def _update_positions(self):
        """
        Retrieves all positions using the REST API.
        """
        product_types: set[str] = {
            await self.product_type_associated_to_trading_pair(trading_pair)
            for trading_pair in self._trading_pairs
        }
        position_sides = {
            "long": PositionSide.LONG,
            "short": PositionSide.SHORT
        }

        for product_type in product_types:
            all_positions_response: Dict[str, Any] = await self._api_get(
                path_url=CONSTANTS.ALL_POSITIONS_ENDPOINT,
                params={
                    "productType": product_type
                },
                is_auth_required=True,
            )
            all_positions_data = all_positions_response["data"]

            for position in all_positions_data:
                symbol = position["symbol"]
                trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol)
                position_side = position_sides[position["holdSide"]]
                unrealized_pnl = Decimal(position["unrealizedPL"])
                entry_price = Decimal(position["openPriceAvg"])
                amount = Decimal(position["total"])
                leverage = Decimal(position["leverage"])

                pos_key = self._perpetual_trading.position_key(
                    trading_pair,
                    position_side
                )

                if amount != s_decimal_0:
                    position_amount = (
                        amount * (
                            Decimal("-1.0")
                            if position_side == PositionSide.SHORT
                            else Decimal("1.0")
                        )
                    )
                    position = Position(
                        trading_pair=trading_pair,
                        position_side=position_side,
                        unrealized_pnl=unrealized_pnl,
                        entry_price=entry_price,
                        amount=position_amount,
                        leverage=leverage,
                    )
                    self._perpetual_trading.set_position(pos_key, position)
                else:
                    self._perpetual_trading.remove_position(pos_key)

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []

        if order.exchange_order_id is not None:
            try:
                all_fills_response = await self._request_order_fills(order=order)
                all_fills_data = all_fills_response["data"]["fillList"]

                for fill_data in all_fills_data:
                    trade_update = self._parse_trade_update(
                        trade_msg=fill_data,
                        tracked_order=order
                    )
                    trade_updates.append(trade_update)
            except IOError as ex:
                if not self._is_request_exception_related_to_time_synchronizer(
                    request_exception=ex
                ):
                    raise

        return trade_updates

    async def _request_order_fills(self, order: InFlightOrder) -> Dict[str, Any]:
        symbol = await self.exchange_symbol_associated_to_pair(order.trading_pair)
        product_type = await self.product_type_associated_to_trading_pair(order.trading_pair)
        order_fills_response = await self._api_get(
            path_url=CONSTANTS.ORDER_FILLS_ENDPOINT,
            params={
                "orderId": order.exchange_order_id,
                "productType": product_type,
                "symbol": symbol,
            },
            is_auth_required=True,
        )
        return order_fills_response

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        try:
            order_status_data = await self._request_order_status_data(tracked_order=tracked_order)
            updated_order_data = order_status_data["data"]

            if len(updated_order_data) == 0:
                raise ValueError(f"Can't parse order status data. Data: {updated_order_data}")

            client_order_id = str(updated_order_data["clientOid"])

            order_update: OrderUpdate = OrderUpdate(
                trading_pair=tracked_order.trading_pair,
                update_timestamp=self.current_timestamp,
                new_state=CONSTANTS.STATE_TYPES[updated_order_data["state"]],
                client_order_id=client_order_id,
                exchange_order_id=updated_order_data["orderId"],
            )

            return order_update

        except IOError as ex:
            if self._is_request_exception_related_to_time_synchronizer(request_exception=ex):
                order_update = OrderUpdate(
                    client_order_id=tracked_order.client_order_id,
                    trading_pair=tracked_order.trading_pair,
                    update_timestamp=self.current_timestamp,
                    new_state=tracked_order.current_state,
                )
            else:
                raise

        return order_update

    async def _request_order_status_data(self, tracked_order: InFlightOrder) -> Dict:
        query_params = {
            "symbol": await self.exchange_symbol_associated_to_pair(tracked_order.trading_pair),
            "productType": await self.product_type_associated_to_trading_pair(
                tracked_order.trading_pair
            )
        }
        if tracked_order.exchange_order_id:
            query_params["orderId"] = tracked_order.exchange_order_id
        else:
            query_params["clientOid"] = tracked_order.client_order_id

        order_detail_response = await self._api_get(
            path_url=CONSTANTS.ORDER_DETAIL_ENDPOINT,
            params=query_params,
            is_auth_required=True,
        )

        return order_detail_response

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair)
        product_type = await self.product_type_associated_to_trading_pair(trading_pair)
        ticker_response = await self._api_get(
            path_url=CONSTANTS.PUBLIC_TICKER_ENDPOINT,
            params={
                "symbol": symbol,
                "productType": product_type
            },
        )

        return float(ticker_response["data"][0]["lastPr"])

    async def set_margin_mode(
        self,
        mode: MarginMode
    ) -> None:
        """
        Change the margin mode of the exchange (cross/isolated)
        """
        margin_mode = CONSTANTS.MARGIN_MODE_TYPES[mode]

        for trading_pair in self.trading_pairs:
            product_type = await self.product_type_associated_to_trading_pair(trading_pair)

            response = await self._api_post(
                path_url=CONSTANTS.SET_MARGIN_MODE_ENDPOINT,
                data={
                    "symbol": await self.exchange_symbol_associated_to_pair(trading_pair),
                    "productType": product_type,
                    "marginMode": margin_mode,
                    "marginCoin": self.get_buy_collateral_token(trading_pair),
                },
                is_auth_required=True,
            )

            if response["code"] != CONSTANTS.RET_CODE_OK:
                self.logger().error(
                    self._formatted_error(
                        response["code"],
                        f"There was an error changing the margin mode ({response['msg']})"
                    )
                )
                return

            self.logger().info(f"Margin mode set to {margin_mode}")

    async def _trading_pair_position_mode_set(
        self,
        mode: PositionMode,
        trading_pair: str
    ) -> Tuple[bool, str]:
        if len(self.account_positions) > 0:
            return False, "Cannot change position because active positions exist"

        try:
            position_mode = CONSTANTS.POSITION_MODE_TYPES[mode]
            product_type = await self.product_type_associated_to_trading_pair(trading_pair)

            response = await self._api_post(
                path_url=CONSTANTS.SET_POSITION_MODE_ENDPOINT,
                data={
                    "productType": product_type,
                    "posMode": position_mode,
                },
                is_auth_required=True,
            )

            if response["code"] != CONSTANTS.RET_CODE_OK:
                return (
                    False,
                    self._formatted_error(response["code"], response["msg"])
                )
        except Exception as exception:
            return (
                False,
                f"There was an error changing the position mode ({exception})"
            )

        return True, ""

    async def _set_trading_pair_leverage(
        self,
        trading_pair: str,
        leverage: int
    ) -> Tuple[bool, str]:
        if len(self.account_positions) > 0:
            return False, "cannot change leverage because active positions exist"

        try:
            product_type = await self.product_type_associated_to_trading_pair(trading_pair)
            symbol = await self.exchange_symbol_associated_to_pair(trading_pair)

            response: Dict[str, Any] = await self._api_post(
                path_url=CONSTANTS.SET_LEVERAGE_ENDPOINT,
                data={
                    "symbol": symbol,
                    "productType": product_type,
                    "marginCoin": self.get_buy_collateral_token(trading_pair),
                    "leverage": str(leverage)
                },
                is_auth_required=True,
            )

            if response["code"] != CONSTANTS.RET_CODE_OK:
                return False, self._formatted_error(response["code"], response["msg"])
        except Exception as exception:
            return (
                False,
                f"There was an error setting the leverage for {trading_pair} ({exception})"
            )

        return True, ""

    async def _fetch_last_fee_payment(self, trading_pair: str) -> Tuple[float, Decimal, Decimal]:
        timestamp, funding_rate, payment = 0, Decimal("-1"), Decimal("-1")

        product_type = await self.product_type_associated_to_trading_pair(trading_pair)
        payment_response: Dict[str, Any] = await self._api_get(
            path_url=CONSTANTS.ACCOUNT_BILLS_ENDPOINT,
            params={
                "productType": product_type,
                "businessType": "contract_settle_fee",
            },
            is_auth_required=True,
        )
        payment_data: Dict[str, Any] = payment_response["data"]["bills"]

        if payment_data:
            last_data = payment_data[0]
            funding_info = self._perpetual_trading._funding_info.get(trading_pair)
            payment: Decimal = Decimal(last_data["amount"])
            funding_rate: Decimal = funding_info.rate if funding_info is not None else Decimal(0)
            timestamp: float = int(last_data["cTime"]) * 1e-3

        return timestamp, funding_rate, payment

    async def _user_stream_event_listener(self):
        async for event_message in self._iter_user_event_queue():
            try:
                channel = event_message["arg"]["channel"]
                data = event_message["data"]

                if channel == CONSTANTS.WS_POSITIONS_ENDPOINT:
                    await self._process_account_position_event(data)
                elif channel == CONSTANTS.WS_ORDERS_ENDPOINT:
                    for order_msg in data:
                        self._process_trade_event_message(order_msg)
                        self._process_order_event_message(order_msg)
                        self._process_balance_update_from_order_event(order_msg)
                elif channel == CONSTANTS.WS_ACCOUNT_ENDPOINT:
                    for wallet_msg in data:
                        self._process_wallet_event_message(wallet_msg)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error in user stream listener loop.")

    async def _process_account_position_event(self, position_entries: List[Dict[str, Any]]):
        """
        Updates position
        :param position_msg: The position event message payload
        """
        all_position_keys = []
        position_sides = {
            "long": PositionSide.LONG,
            "short": PositionSide.SHORT
        }

        for position in position_entries:
            symbol = position["instId"]
            trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol)
            position_side = position_sides[position["holdSide"]]
            entry_price = Decimal(position["openPriceAvg"])
            amount = Decimal(position["total"])
            leverage = Decimal(position["leverage"])
            unrealized_pnl = Decimal(position["unrealizedPL"])

            pos_key = self._perpetual_trading.position_key(trading_pair, position_side)
            all_position_keys.append(pos_key)

            if amount != s_decimal_0:
                position_amount = (
                    amount * (
                        Decimal("-1.0")
                        if position_side == PositionSide.SHORT
                        else Decimal("1.0")
                    )
                )
                position = Position(
                    trading_pair=trading_pair,
                    position_side=position_side,
                    unrealized_pnl=unrealized_pnl,
                    entry_price=entry_price,
                    amount=position_amount,
                    leverage=leverage,
                )
                self._perpetual_trading.set_position(pos_key, position)
            else:
                self._perpetual_trading.remove_position(pos_key)

        # Bitget sends position events as snapshots.
        # If a position is closed it is just not included in the snapshot
        position_keys = list(self.account_positions.keys())
        positions_to_remove = (
            position_key
            for position_key in position_keys
            if position_key not in all_position_keys
        )
        for position_key in positions_to_remove:
            self._perpetual_trading.remove_position(position_key)

    def _process_order_event_message(self, order_msg: Dict[str, Any]):
        """
        Updates in-flight order and triggers cancellation or failure event if needed.

        :param order_msg: The order event message payload
        """
        order_status = CONSTANTS.STATE_TYPES[order_msg["status"]]
        client_order_id = str(order_msg["clientOid"])
        updatable_order = self._order_tracker.all_updatable_orders.get(client_order_id)

        if updatable_order is not None:
            new_order_update: OrderUpdate = OrderUpdate(
                trading_pair=updatable_order.trading_pair,
                update_timestamp=self.current_timestamp,
                new_state=order_status,
                client_order_id=client_order_id,
                exchange_order_id=order_msg["orderId"],
            )
            self._order_tracker.process_order_update(new_order_update)

    def _process_balance_update_from_order_event(self, order_msg: Dict[str, Any]):
        order_status = CONSTANTS.STATE_TYPES[order_msg["status"]]
        symbol = order_msg["marginCoin"]
        states_to_consider = [OrderState.OPEN, OrderState.CANCELED]
        order_amount = Decimal(order_msg["size"])
        order_price = Decimal(order_msg["price"])
        margin_amount = (order_amount * order_price) / Decimal(order_msg["leverage"])
        is_opening = order_msg["tradeSide"] in [
            "open",
            "buy_single",
            "sell_single",
        ]

        if (
            symbol in self._account_available_balances
            and order_status in states_to_consider
            and is_opening
        ):
            multiplier = Decimal(-1) if order_status == OrderState.OPEN else Decimal(1)
            self._account_available_balances[symbol] += margin_amount * multiplier

    def _process_trade_event_message(self, trade_msg: Dict[str, Any]):
        """
        Updates in-flight order and trigger order filled event for trade message received.
        Triggers order completed event if the total executed amount equals to the specified order amount.

        :param trade_msg: The trade event message payload
        """

        client_order_id = str(trade_msg["clientOid"])
        fillable_order = self._order_tracker.all_fillable_orders.get(client_order_id)

        if fillable_order and "tradeId" in trade_msg:
            trade_update = self._parse_websocket_trade_update(
                trade_msg=trade_msg,
                tracked_order=fillable_order
            )
            if trade_update:
                self._order_tracker.process_trade_update(trade_update)

    def _parse_websocket_trade_update(
        self,
        trade_msg: Dict,
        tracked_order: InFlightOrder
    ) -> TradeUpdate:
        trade_id: str = trade_msg["tradeId"]

        if trade_id is not None:
            trade_id = str(trade_id)
            fee_asset = trade_msg["fillFeeCoin"]
            fee_amount = -Decimal(trade_msg["fillFee"])
            position_actions = {
                "open": PositionAction.OPEN,
                "close": PositionAction.CLOSE,
            }
            position_action = position_actions.get(trade_msg["tradeSide"], PositionAction.NIL)
            flat_fees = (
                [] if fee_amount == Decimal("0")
                else [TokenAmount(amount=fee_amount, token=fee_asset)]
            )

            fee = TradeFeeBase.new_perpetual_fee(
                fee_schema=self.trade_fee_schema(),
                position_action=position_action,
                percent_token=fee_asset,
                flat_fees=flat_fees,
            )

            exec_price = (
                Decimal(trade_msg["fillPrice"])
                if "fillPrice" in trade_msg
                else Decimal(trade_msg["price"])
            )
            exec_time = int(trade_msg["fillTime"]) * 1e-3

            trade_update: TradeUpdate = TradeUpdate(
                trade_id=trade_id,
                client_order_id=tracked_order.client_order_id,
                exchange_order_id=str(trade_msg["orderId"]),
                trading_pair=tracked_order.trading_pair,
                fill_timestamp=exec_time,
                fill_price=exec_price,
                fill_base_amount=Decimal(trade_msg["baseVolume"]),
                fill_quote_amount=exec_price * Decimal(trade_msg["baseVolume"]),
                fee=fee,
            )

            return trade_update

    def _parse_trade_update(self, trade_msg: Dict, tracked_order: InFlightOrder) -> TradeUpdate:
        fee_detail = trade_msg["feeDetail"][0]
        fee_asset = fee_detail["feeCoin"]
        fee_amount = abs(Decimal((
            fee_detail["totalDeductionFee"]
            if fee_detail.get("deduction") == "yes"
            else fee_detail["totalFee"]
        )))
        position_actions = {
            "open": PositionAction.OPEN,
            "close": PositionAction.CLOSE,
        }
        position_action = position_actions.get(trade_msg["tradeSide"], PositionAction.NIL)
        flat_fees = (
            [] if fee_amount == Decimal("0")
            else [TokenAmount(amount=fee_amount, token=fee_asset)]
        )

        fee = TradeFeeBase.new_perpetual_fee(
            fee_schema=self.trade_fee_schema(),
            position_action=position_action,
            percent_token=fee_asset,
            flat_fees=flat_fees,
        )

        exec_price = Decimal(trade_msg["price"])
        exec_time = int(trade_msg["cTime"]) * 1e-3

        trade_update: TradeUpdate = TradeUpdate(
            trade_id=trade_msg["tradeId"],
            client_order_id=tracked_order.client_order_id,
            exchange_order_id=trade_msg["orderId"],
            trading_pair=tracked_order.trading_pair,
            fill_timestamp=exec_time,
            fill_price=exec_price,
            fill_base_amount=Decimal(trade_msg["baseVolume"]),
            fill_quote_amount=exec_price * Decimal(trade_msg["baseVolume"]),
            fee=fee,
        )

        return trade_update

    def _process_wallet_event_message(self, wallet_msg: Dict[str, Any]):
        """
        Updates account balances.
        :param wallet_msg: The account balance update message payload
        """
        symbol = wallet_msg["marginCoin"]
        available = Decimal(wallet_msg["maxOpenPosAvailable"])
        total = Decimal(wallet_msg["equity"])

        self._account_balances[symbol] = total
        self._account_available_balances[symbol] = available

    async def _make_trading_pairs_request(self) -> Any:
        all_exchange_info: List[Dict[str, Any]] = []

        for product_type in CONSTANTS.ALL_PRODUCT_TYPES:
            exchange_info = await self._api_get(
                path_url=self.trading_pairs_request_path,
                params={
                    "productType": product_type
                }
            )
            all_exchange_info.extend(exchange_info["data"])

        return all_exchange_info

    async def _make_trading_rules_request(self) -> Any:
        return await self._make_trading_pairs_request()

    def _initialize_trading_pair_symbols_from_exchange_info(
        self,
        exchange_info: List[Dict[str, Any]]
    ) -> None:
        mapping = bidict()
        for symbol_data in exchange_info:
            if bitget_perpetual_utils.is_exchange_information_valid(exchange_info=symbol_data):
                try:
                    symbol = symbol_data["symbol"]
                    base = symbol_data["baseCoin"]
                    quote = symbol_data["quoteCoin"]
                    trading_pair = combine_to_hb_trading_pair(base, quote)
                    mapping[symbol] = trading_pair
                except Exception as exception:
                    self.logger().error(
                        f"There was an error parsing a trading pair information ({exception}). Symbol: {symbol}. Trading pair: {trading_pair}"
                    )
        self._set_trading_pair_symbol_map(mapping)

    async def _format_trading_rules(
        self,
        exchange_info_dict: Dict[str, List[Dict[str, Any]]]
    ) -> List[TradingRule]:
        """
        Converts JSON API response into a local dictionary of trading rules.

        :param instrument_info_dict: The JSON API response.

        :returns: A dictionary of trading pair to its respective TradingRule.
        """
        trading_rules = []
        for rule in exchange_info_dict:
            if bitget_perpetual_utils.is_exchange_information_valid(exchange_info=rule):
                try:
                    trading_pair = await self.trading_pair_associated_to_exchange_symbol(
                        symbol=rule["symbol"]
                    )
                    max_order_size = Decimal(rule["maxOrderQty"]) if rule["maxOrderQty"] else None
                    margin_coin = rule["supportMarginCoins"][0]

                    trading_rules.append(
                        TradingRule(
                            trading_pair=trading_pair,
                            min_order_value=Decimal(rule["minTradeUSDT"]),
                            max_order_size=max_order_size,
                            min_order_size=Decimal(rule["minTradeNum"]),
                            min_price_increment=Decimal(f"1e-{int(rule['pricePlace'])}"),
                            min_base_amount_increment=Decimal(rule["sizeMultiplier"]),
                            buy_order_collateral_token=margin_coin,
                            sell_order_collateral_token=margin_coin,
                        )
                    )
                except Exception:
                    self.logger().exception(
                        f"Error parsing the trading pair rule: {rule}. Skipping."
                    )

        return trading_rules
